"""
    Optimizer.StrategyEditor.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""
import os
from tksheet import Sheet
from molass_legacy.KekLib.OurTkinter import Tk, Dialog
from molass_legacy.KekLib.ImportUtils import import_module_from_path

class StrategyEditor(Dialog):
    def __init__(self, parent):
        self.parent = parent
        Dialog.__init__(self, parent, "Strategy Editor", visible=False)

    def show(self):
        self._show()
    
    def body(self, body_frame):
        num_components = self.parent.get_num_peaks()
        model_name = self.parent.get_model_name()
        module_name = model_name.title() + "Sheet"
        module_path = os.path.join(os.path.join(os.path.dirname(__file__), "Strategies"), module_name + ".py")
        module = import_module_from_path(module_name, module_path)
        data_list = module.get_data_list(num_components)
        self.sheet = Sheet(body_frame, width=800, height=600, data=data_list)
        self.sheet.headers(["parameter name", "first", "middle", "last"])
        self.sheet.pack()


"""
    QuickAnalysis.PeakUtils.py

    Copyright (c) 2021-2023, SAXS Team, KEK-PF
"""
from .ModeledPeaks import *         # for backward compatibility

def demo_modeled_peaks(in_folder):
    from molass_legacy._MOLASS.SerialSettings import set_setting
    from molass_legacy.Batch.StandardProcedure import StandardProcedure
    from molass_legacy.Trimming.PreliminaryRecognition import PreliminaryRecognition
    from molass_legacy.Elution.CurveUtils import simple_plot

    sp = StandardProcedure()
    sd_ = sp.load_old_way(in_folder)
    pre_recog = PreliminaryRecognition(sd_)
    sd = sd_._get_analysis_copy_impl(pre_recog)
    uv_x, uv_y, xr_x, xr_y, details = get_curve_xy_impl(sd, baseline_type=1)
    uv_peaks, xr_peaks, a, b = get_modeled_peaks_impl(uv_x, uv_y, xr_x, xr_y, num_peaks=5, debug=True)

def demo_modeled_peaks_dialog(root, in_folder):
    from molass_legacy._MOLASS.SerialSettings import set_setting
    from molass_legacy.Batch.StandardProcedure import StandardProcedure
    from molass_legacy.Trimming.PreliminaryRecognition import PreliminaryRecognition
    from molass_legacy.Elution.CurveUtils import simple_plot
    from importlib import reload
    import QuickAnalysis.ModeledPeaksTester
    reload(QuickAnalysis.ModeledPeaksTester)
    from molass_legacy.QuickAnalysis.ModeledPeaksTester import TestDialog

    sp = StandardProcedure()
    sd_ = sp.load_old_way(in_folder)
    pre_recog = PreliminaryRecognition(sd_)
    sd = sd_._get_analysis_copy_impl(pre_recog)
    dialog = TestDialog(root, sd)
    dialog.show()

def guess_peak_params(xr_curve, pos, j, adj_params):
    x = xr_curve.x
    y = xr_curve.y
    spline = xr_curve.spline
    h, m, s, t = adj_params

    def objective(p):
        n_h, n_s, n_t, a_h, a_s, a_t = p
        ty = egh(x, n_h, pos, n_s, n_t) + egh(x, a_h, m, a_s, a_t)
        ry = ty - y
        return np.sum(ry[x < m]**2)

    i_h = spline(pos)
    ret = minimize(objective, (i_h, 0.5*s, 0, h, s, 0))
    n_h, n_s, n_t, a_h, a_s, a_t = ret.x
    return [(n_h, pos, n_s, n_t), (a_h, m, a_s, a_t)]

class DataFilter:
    def __init__(self, names):
        self.names = names

    def ok(self, in_folder):
        for name in self.names:
            if in_folder.find(name) >= 0:
                return True
        return False

class ModelCurve:
    def __init__(self, h, mu, sigma, tau):
        self.params = (h, mu, sigma, tau)

    def __call__(self, x):
        if np.isscalar(x):
            return egh(np.array([x]), *self.params)[0]
        else:
            return egh(x, *self.params)



# coding: utf-8
"""
    FewerPointMapper.py

    Copyright (c) 2021-2022, SAXS Team, KEK-PF
"""
import logging
import numpy as np
from scipy import stats
from scipy.interpolate import UnivariateSpline
from matplotlib.gridspec import GridSpec
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy.Trimming.PreliminaryRecognition import PreliminaryRecognition
from molass_legacy.Elution.CurveUtils import simple_plot
from molass_legacy.Peaks.RobustCurve import RobustCurve, plot_curve, gaussian
from molass_legacy.Trimming.MatchingPeaks import compute_matching_indeces

class FewerPointMapper:
    def __init__(self, rb_curve1, rb_curve2):
        self.curves = [rb_curve1, rb_curve2]
        v = np.array([pt[0] for pt in rb_curve1.get_peak_tops()])
        u = np.array([pt[0] for pt in rb_curve2.get_peak_tops()])
        scale = len(rb_curve2.x)/len(rb_curve1.x)
        indeces = compute_matching_indeces(len(rb_curve1.x), v*scale, u)
        j_ = []
        i_ = []
        for j, i in zip(indeces[0], indeces[1]):
            if j is None or i is None:
                continue
            j_.append(j)
            i_.append(i)
        slope, intercept, r_value, p_value, std_err = stats.linregress(u[i_], v[j_])
        self.params = (slope, intercept)
        self.spline = UnivariateSpline(rb_curve1.x, rb_curve1.y, s=0, ext=3)
        self.matching_indeces = np.array([j_, i_])

    def get_params(self):
        return self.params

    def get_simply_mapped_y(self, u):
        slope, intercept = self.params
        return self.spline(u*slope + intercept)

    def get_scale_mapped_y(self, u):
        simple_y  = self.get_simply_mapped_y(u)

        peak_model_list = []
        for k, curve in enumerate(self.curves):
            models = curve.get_peak_models()
            matched = []
            for i in self.matching_indeces[k,:]:
                matched.append(models[i])
            peak_model_list.append(matched)

        slope, intercept = self.params
        mapping_params = []
        tw = np.zeros(len(u))
        for model1, model2 in zip(*peak_model_list):
            if len(model1.params) == 3:
                h1, m1, s1 = model1.params
                h2, m2, s2 = model2.params
                m1_ = (m1 - intercept)/slope
                s1_ = s1/slope
                w = model2.func(u, h2, m1_, s1_)
            elif len(model1.params) == 4:
                h1, m1, s1, t1 = model1.params
                h2, m2, s2, t2 = model2.params
                m1_ = (m1 - intercept)/slope
                s1_ = s1/slope
                t1_ = t1/slope
                w = model2.func(u, h2, m1_, s1_, t1_)
            else:
                assert False
            mapping_params.append((w, h2/h1))
            tw += w

        mapped_y = np.zeros(len(u))
        for w, scale in mapping_params:
            mapped_y += simple_y * w/tw *scale

        return mapped_y

def demo(sd, peak_model_class, scale_mapped=True):
    from DataUtils import get_in_folder
    from molass_legacy.Peaks.RobustCurve import LIFTED_BASE, MINIMIZE_METHOD

    peak_model_name = peak_model_class.name
    if peak_model_name == 'Hyperbola':
        scale_mapped = False

    uv_curve = sd.get_uv_curve()
    xr_curve = sd.get_xray_curve()

    rb_curves = []
    for curve in [uv_curve, xr_curve]:
        x = curve.x
        y = curve.y
        rbc = RobustCurve(x, y, peak_model=peak_model_class)
        rb_curves.append(rbc)

    fpm = FewerPointMapper(*rb_curves)

    plt.push()
    fig, (ax1, ax2, ax3) = plt.subplots(ncols=3, figsize=(21,7))

    text_args = []
    if LIFTED_BASE:
        text_args.append("lifted base")
    method = "default method" if MINIMIZE_METHOD is None else MINIMIZE_METHOD
    text_args.append(method)
    paren_text = "(%s)" % ', '.join(text_args)

    fig.suptitle("Fewer Point Mapping with Locally fitted %s for %s  %s" % (peak_model_name, get_in_folder(), paren_text), fontsize=20)
    ax1.set_title("UV Elution", fontsize=16)
    ax2.set_title("Xray Elution", fontsize=16)
    how = 'Scale' if scale_mapped else 'Simply'
    ax3.set_title("%s mapped Elution" % how, fontsize=16)

    plot_curve(ax1, rb_curves[0], color='C0')
    plot_curve(ax2, rb_curves[1], color='C1')

    x = xr_curve.x
    y = xr_curve.y


    if scale_mapped:
        my = fpm.get_scale_mapped_y(x)
        ax_ = ax3
    else:
        my = fpm.get_simply_mapped_y(x)
        axt = ax3.twinx()
        axt.grid(False)
        ax_ = axt

    ax3.plot(x, y, color='C1')
    ax_.plot(x, my, color='C0')

    for ax in [ax1, ax2, ax3]:
        ax.legend()

    fig.tight_layout()
    plt.show()
    plt.pop()



"""
    Implementation of Matrix Factorization in Python
    https://github.com/albertauyeung/matrix-factorization-in-python
"""
import numpy as np


class MF():
    
    def __init__(self, R, K, alpha, beta, iterations):
        """
        Perform matrix factorization to predict empty
        entries in a matrix.
        
        Arguments
        - R (ndarray)   : user-item rating matrix
        - K (int)       : number of latent dimensions
        - alpha (float) : learning rate
        - beta (float)  : regularization parameter
        """
        
        self.R = R
        self.num_users, self.num_items = R.shape
        self.K = K
        self.alpha = alpha
        self.beta = beta
        self.iterations = iterations

    def train(self):
        # Initialize user and item latent feature matrice
        self.P = np.random.normal(scale=1./self.K, size=(self.num_users, self.K))
        self.Q = np.random.normal(scale=1./self.K, size=(self.num_items, self.K))
        
        # Initialize the biases
        self.b_u = np.zeros(self.num_users)
        self.b_i = np.zeros(self.num_items)
        self.b = np.mean(self.R[np.where(self.R != 0)])
        
        # Create a list of training samples
        self.samples = [
            (i, j, self.R[i, j])
            for i in range(self.num_users)
            for j in range(self.num_items)
            if self.R[i, j] > 0
        ]
        
        # Perform stochastic gradient descent for number of iterations
        training_process = []
        for i in range(self.iterations):
            np.random.shuffle(self.samples)
            self.sgd()
            mse = self.mse()
            training_process.append((i, mse))
            if (i+1) % 10 == 0:
                print("Iteration: %d ; error = %.4f" % (i+1, mse))
        
        return training_process

    def mse(self):
        """
        A function to compute the total mean square error
        """
        xs, ys = self.R.nonzero()
        predicted = self.full_matrix()
        error = 0
        for x, y in zip(xs, ys):
            error += pow(self.R[x, y] - predicted[x, y], 2)
        return np.sqrt(error)

    def sgd(self):
        """
        Perform stochastic graident descent
        """
        for i, j, r in self.samples:
            # Computer prediction and error
            prediction = self.get_rating(i, j)
            e = (r - prediction)
            
            # Update biases
            self.b_u[i] += self.alpha * (e - self.beta * self.b_u[i])
            self.b_i[j] += self.alpha * (e - self.beta * self.b_i[j])
            
            # Create copy of row of P since we need to update it but use older values for update on Q
            P_i = self.P[i, :][:]
            
            # Update user and item latent feature matrices
            self.P[i, :] += self.alpha * (e * self.Q[j, :] - self.beta * self.P[i,:])
            self.Q[j, :] += self.alpha * (e * P_i - self.beta * self.Q[j,:])

    def get_rating(self, i, j):
        """
        Get the predicted rating of user i and item j
        """
        prediction = self.b + self.b_u[i] + self.b_i[j] + self.P[i, :].dot(self.Q[j, :].T)
        return prediction
    
    def full_matrix(self):
        """
        Computer the full matrix using the resultant biases, P and Q
        """
        return self.b + self.b_u[:,np.newaxis] + self.b_i[np.newaxis:,] + self.P.dot(self.Q.T)



"""Inverse transform sampling, for sampling from arbitrary probability density
functions."""

import numpy as np
from scipy.integrate import quad
from scipy.optimize import root
from numpy.random import uniform

from numpy.polynomial.chebyshev import chebfit, chebval, chebint

# Functions to check if a numerical quadrature is convergent

_MESSAGE = "The integral is probably divergent, or slowly convergent."

def _convergent(quadrature):
    if len(quadrature)>3 and quadrature[3] == _MESSAGE:
        return False
    else:
        return True


def normalize(pdf, lower_bd=-np.inf, upper_bd=np.inf, vectorize=False):
    """Normalize a non-normalized PDF.
    
    Parameters
    ----------
    pdf : function, float -> float
        The probability density function (not necessarily normalized). Must take
        floats or ints as input, and return floats as an output.
    lower_bd : float
        Lower bound of the support of the pdf. This parameter allows one to
        manually establish cutoffs for the density.
    upper_bd : float
        Upper bound of the support of the pdf.
    vectorize: boolean
        Vectorize the function. This slows down function calls, and so is
        generally set to False.
    
    Returns
    -------
    pdf_norm : function
        Function with same signature as pdf, but normalized so that the integral
        between lower_bd and upper_bd is close to 1. Maps nicely over iterables.
    """
    if lower_bd >= upper_bd:
        raise ValueError('Lower bound must be less than upper bound.')
    quadrature = quad(pdf, lower_bd, upper_bd,full_output=1)
    if not _convergent(quadrature):
        raise ValueError('PDF integral likely divergent.')
    A = quadrature[0]
    pdf_normed = lambda x: pdf(x)/A if lower_bd <= x <= upper_bd else 0
    if vectorize:
        def pdf_vectorized(x):
            try:
                return pdf_normed(x)
            except ValueError:
                return np.array([pdf_normed(xi) for xi in x])
        return pdf_vectorized
    else:
        return pdf_normed


def get_cdf(pdf, lower_bd=-np.inf, upper_bd=np.inf):
    """Generate a CDF from a (possibly not normalized) pdf.
    
    Parameters
    ----------
    pdf : function, float -> float
        The probability density function (not necessarily normalized). Must take
        floats or ints as input, and return floats as an output.
    lower_bd : float
        Lower bound of the support of the pdf. This parameter allows one to
        manually establish cutoffs for the density.
    upper_bd : float
        Upper bound of the support of the pdf.
    
    Returns
    -------
    cdf : function
        The cumulative density function of the (normalized version of the)
        provided pdf. Will return a float if provided with a float or int; will
        return a numpy array if provided with an iterable.

    """
    pdf_norm = normalize(pdf, lower_bd, upper_bd)
    
    def cdf_number(x):
        "Numerical cdf"""
        if x < lower_bd:
            return 0.0
        elif x > upper_bd:
            return 1.0
        else:
            return quad(pdf_norm,lower_bd,x)[0]

    def cdf_vector(x):
        try:
            return np.array([cdf_number(xi) for xi in x])
        except AttributeError:
            return cdf_number(x)

    return cdf_vector

############################################
### CHEBYSHEV APPROXIMATION OF PDF & CDF ###
############################################

# This section follows the work of Olver & Townsend (2013), who suggest using
# Chebyshev polynomials to approximate the PDF. This allows for trivial
# construction of the CDF, as these polynomials can be integrated in closed
# form.


def _chebnodes(a,b,n):
    """Chebyshev nodes of rank n on interal [a,b]."""
    if not a < b:
        raise ValueError('Lower bound must be less than upper bound.')
    return np.array([1/2*((a+b)+(b-a)*np.cos((2*k-1)*np.pi/(2*n))) for k in range(1,n+1)])


def adaptive_chebfit(pdf, lower_bd, upper_bd, eps=10**(-15)):
    """Fit a chebyshev polynomial, increasing sampling rate until coefficient
    tail falls below provided tolerance.
    
    Parameters
    ----------
    pdf : function, float -> float
        The probability density function (not necessarily normalized). Must take
        floats or ints as input, and return floats as an output.
    lower_bd : float
        Lower bound of the support of the pdf. This parameter allows one to
        manually establish cutoffs for the density.
    upper_bd : float
        Upper bound of the support of the pdf.
    eps: float
        Error tolerance of Chebyshev polynomial fit of PDF.

    Returns
    -------
    x : array
        The nodes at which the polynomial interpolation takes place. These are
        adaptively chosen based on the provided tolerance.
    coeffs : array
        Coefficients in Chebyshev approximation of the PDF.

    Notes
    -----
    This fit defines the "error" as the magnitude of the tail of the Chebyshev
    coefficients. Computing the true error (i.e. discrepancy between the PDF and
    it's approximant) would be much slower, so we avoid it and use this rough
    approximation in its place.

    """
    i = 4
    error = eps+1 # so that it runs the first time through
    while error > eps:
        n = 2**i + 1
        x = _chebnodes(lower_bd,upper_bd,n)
        y = pdf(x)
        coeffs = chebfit(x,y,n-1)
        error = max(np.abs(coeffs[-5:]))
        i += 1
    return x,coeffs
    

def chebcdf(pdf, lower_bd, upper_bd, eps=10**(-15)):
    """Get Chebyshev approximation of the CDF.
    
    Parameters
    ----------
    pdf : function, float -> float
        The probability density function (not necessarily normalized). Must take
        floats or ints as input, and return floats as an output.
    lower_bd : float
        Lower bound of the support of the pdf. This parameter allows one to
        manually establish cutoffs for the density.
    upper_bd : float
        Upper bound of the support of the pdf.
    eps: float
        Error tolerance of Chebyshev polynomial fit of PDF.
    
    Returns
    -------
    cdf : function
        The cumulative density function of the (normalized version of the)
        provided pdf. The function cdf() takes an iterable of floats or doubles
        as an argument, and returns an iterable of floats of the same length.
    """
    if not (np.isfinite(lower_bd) and np.isfinite(upper_bd)):
        raise ValueError('Bounds must be finite.')
    if not lower_bd < upper_bd:
        raise ValueError('Lower bound must be less than upper bound.')

    x,coeffs = adaptive_chebfit(pdf, lower_bd, upper_bd, eps)
    int_coeffs = chebint(coeffs)
    # offset and scale so that it goes from 0 to 1, i.e. is a true CDF.
    offset = chebval(lower_bd, int_coeffs)
    scale = chebval(upper_bd, int_coeffs) - chebval(lower_bd, int_coeffs)
    cdf = lambda x: (chebval(x,int_coeffs) - offset) / scale
    return cdf

# Sample via root-finding on CDF

def sample(pdf, num_samples, lower_bd=-np.inf, upper_bd=np.inf, guess=0, chebyshev=False):
    """Sample from an arbitrary, unnormalized PDF.
    
    Parameters
    ----------
    pdf : function, float -> float
        The probability density function (not necessarily normalized). Must take
        floats or ints as input, and return floats as an output.
    num_samples : int
        The number of samples to be generated.
    lower_bd : float
        Lower bound of the support of the pdf. This parameter allows one to
        manually establish cutoffs for the density.
    upper_bd : float
        Upper bound of the support of the pdf.
    guess : float or int
        Initial guess for the numerical solver to use when inverting the CDF.
    chebyshev: Boolean, optional (default=False)
        If True, then the CDF is approximated using Chebyshev polynomials.
    
    Returns
    -------
    samples : numpy array
        An array of samples from the provided PDF, with support between lower_bd
        and upper_bd. 

    Notes
    -----
    For a unimodal distribution, the mode is a good choice for the parameter
    guess. Any number for which the CDF is not extremely close to 0 or 1 should
    be acceptable. If the cdf(guess) is near 1 or 0, then its derivative is near 0,
    and so the numerical root finder will be very slow to converge.

    This sampling technique is slow (~3 ms/sample for a unit normal with initial
    guess of 0), since we re-integrate to get the CDF at every iteration of the
    numerical root-finder. This is improved somewhat by using Chebyshev
    approximations of the CDF, but the sampling rate is still prohibitively slow
    for >100k samples.

    """
    if chebyshev:
        if not (np.isfinite(lower_bd) and np.isfinite(upper_bd)):
            raise ValueError('Bounds must be finite for Chebyshev approximation of CDF.')
        cdf = chebcdf(pdf, lower_bd, upper_bd)
    else:
        cdf = get_cdf(pdf, lower_bd, upper_bd)
    seeds = uniform(0, 1, num_samples)
    samples = []
    for seed in seeds:
        shifted = lambda x: cdf(x)-seed
        soln = root(shifted,guess)
        samples.append(soln.x[0])
    return np.array(samples)



"""
    RangeEditorDialog.py

    Copyright (c) 2019-2024, SAXS Team, KEK-PF
"""
import os
import logging
from molass_legacy.KekLib.OurTkinter import Tk, Dialog
from molass_legacy.KekLib.TkSupplements import set_icon
from molass_legacy.KekLib.TkUtils import is_low_resolution
from molass_legacy.Decomposer.DecompUtils import CorrectedBaseline
from .RangeEditorFrame import RangeEditorFrame
from molass_legacy.Models.ElutionCurveModels     import EGHA, EMGA
from molass_legacy._MOLASS.SerialSettings         import get_setting, set_setting

class RangeEditorDialog(Dialog):
    def __init__( self, parent, title, sd, mapper, judge_holder, scrollable=False ):
        self.logger  = logging.getLogger( __name__ )
        self.parent = parent
        self.title_ = title
        self.sd     = sd
        self.serial_data = sd   # for compatibility with AnalyzerDialog for Global Fitting
        self.mapper = mapper
        self.judge_holder = judge_holder
        self.ready = False
        self.scrollable = scrollable or is_low_resolution()
        self.canceled = False
        self.applied = False
        self.init_use_elution_models = get_setting('use_elution_models')
        set_setting('use_elution_models', 1)    # required to preview in this dialog

    def is_ready( self ):
        return self.ready

    def prepare( self, counter=None):
        self.parent.config( cursor='wait' )
        self.counter = counter
        self.parent.update()
        Dialog.__init__( self, self.parent, self.title_, visible=False, block=False )
        self.parent.config( cursor='' )
        self.ready = True

    def cancel(self):
        # overiding cancel to cleanup self.fig
        # because the call to the destructor __del__ seems to be delayed
        self.editor_frame.close_figs()
        print("RangeEditorDialog: closed figs")
        Dialog.cancel(self)

    def user_cancel(self):
        # this method is provided aside from "cancel" to avoid confusion.
        set_setting('use_elution_models', self.init_use_elution_models)
        self.cancel()

    def show(self, counter=None):
        if not self.ready:
            # for backward compatibility
            self.prepare(counter=counter)

        self._show()

    def body( self, default_body_frame ):   # overrides parent class method
        set_icon( self )

        if self.scrollable:
            from molass_legacy.KekLib.ScrolledFrame import ScrolledFrame
            self.scrolled_frame = ScrolledFrame(default_body_frame)
            # scrolled_frame.pack( fill=Tk.BOTH, expand=1 )
            self.scrolled_frame.pack(anchor=Tk.N)
            body_frame = self.scrolled_frame.interior
        else:
            self.scrolled_frame = None
            body_frame = default_body_frame

        corbase_info = CorrectedBaseline(self.sd, self.mapper)      # necessary?
        self.editor_frame = RangeEditorFrame(body_frame, self, self.sd, self.mapper, corbase_info)
        self.editor_frame.pack()

    def buttonbox(self, devel=False):
        if devel:
            from importlib import reload
            import molass_legacy.Extrapolation.PreviewButtonFrame
            reload(molass_legacy.Extrapolation.PreviewButtonFrame)
        from molass_legacy.Extrapolation.PreviewButtonFrame import PreviewButtonFrame

        box = Tk.Frame(self)
        box.pack(fill=Tk.X)

        lower_frame = Tk.Frame(box)
        lower_frame.pack(fill=Tk.X)

        additional_box = Tk.Frame(lower_frame)
        additional_box.pack(side=Tk.RIGHT, anchor=Tk.N + Tk.E)
        standard_box = Tk.Frame(lower_frame)
        standard_box.pack(anchor=Tk.CENTER)

        self.guide_frame = Tk.Frame(box)
        self.guide_frame.pack(fill=Tk.X)
        ok_cancel_frame = Tk.Frame(box)
        ok_cancel_frame.pack()

        self.preview_frame = PreviewButtonFrame(standard_box,
                                    dialog=self, use_elution_models=False, judge_holder=self.judge_holder,
                                    conc_opts=True, editor=self.editor_frame)
        self.preview_frame.grid(row=0, column=0, columnspan=2, padx=5, pady=5)

        from molass_legacy.QuickAnalysis.AnalysisGuide import get_data_correction_state_text
        text = "Showing data processed with " + get_data_correction_state_text()
        guide_label = Tk.Label(self.guide_frame, text=text, bg='white')
        guide_label.pack(fill=Tk.X, pady=10)
        self.guide_message = guide_label

        w = Tk.Button(ok_cancel_frame, text="OK", width=10, command=self.ok, default=Tk.ACTIVE)
        w.grid(row=1, column=0, padx=5, pady=5)
        self.ok_button = w

        space = Tk.Frame(ok_cancel_frame, width=160)
        space.grid(row=0, column=1)

        w = Tk.Button(ok_cancel_frame, text="Cancel", width=10, command=self.user_cancel)
        w.grid(row=1, column=2, padx=5, pady=5)
        self.cancel_button = w

        if False:
            w = Tk.Button(additional_box, text="Reset from Log", width=12, command=self.show_reset_from_log_dialog, state=Tk.DISABLED)
            w.pack(padx=10, pady=5)
            self.restore_button = w

        # self.bind("<Return>", self.ok)
        self.bind("<Escape>", self.user_cancel)

    def get_conc_depend(self):
        return self.preview_frame.get_conc_depend()

    def apply( self ):
        self.editor_frame.apply()
        self.preview_frame.update_settings()        # concentration_datatype will be set_setting-ed here
        self.applied = True
        set_setting('use_elution_models', 0)
        set_setting('matrix_formulation', 1)
        self.try_image_save()

    def try_image_save( self ):
        # TODO: move this code to Tester.py

        test_pattern = get_setting('test_pattern')
        if test_pattern is None or test_pattern < 7:
            return

        fig_type = self.get_current_frame().elution_fig_type
        data_type = '_xray' if fig_type == 0 else '_uv'
        decomp_image_folder = get_setting('decomp_image_folder' + data_type )
        if decomp_image_folder is not None and os.path.exists(decomp_image_folder):
            self.save_the_figure( decomp_image_folder, get_setting('analysis_name') )

    def save_the_figure(self, folder, analysis_name):
        from pyautogui import screenshot
        from molass_legacy.KekLib.TkUtils import split_geometry

        filename = analysis_name.replace( 'analysis', 'figure' ) + '.png'
        path = os.path.join( folder, filename )
        w, h, x, y = split_geometry(self.geometry())
        # print( path, w, h, x, y  )
        screenshot( path, region=(x, y, w, h) ) # doesn't support multi-screen

    def make_range_info(self, concentration_datatype):
        set_setting('concentration_datatype', concentration_datatype)
        return self.editor_frame.make_range_info(concentration_datatype)

    def show_reset_from_log_dialog(self):
        # TODO:
        from molass_legacy.Decomposer.DecompResetDialog import DecompResetDialog
        self.get_current_frame().log_info_for_reset(modification="currently active")
        dialog = DecompResetDialog(self)
        dialog.show()



# coding: utf-8
"""
    QmmWindowSetting.py

    Copyright (c) 2020, SAXS Team, KEK-PF
"""
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from matplotlib.widgets import SpanSelector
from matplotlib.patches import Rectangle
from bisect import bisect_right
from molass_legacy.KekLib.OurTkinter import Tk, Dialog
from molass_legacy.KekLib.OurMatplotlib import NavigationToolbar
from molass_legacy._MOLASS.SerialSettings import get_setting, set_setting

class QmmWindowSetting(Dialog):
    def __init__(self, parent, outline_fig_frame=None):
        self.outline_fig_frame = outline_fig_frame
        if outline_fig_frame is None:
            self.cs = None
            self.restrict = None
        else:
            self.cs = outline_fig_frame.pre_recog.cs
            self.restrict = outline_fig_frame.xray_elution_restrict
        self.window_patch = None
        Dialog.__init__(self, parent, title="QMM Window Setting", visible=False)

    def show(self):
        # self.after(0, self._adjust_geometry)
        self._show()

    def body(self, body_frame):

        cframe = Tk.Frame(body_frame)
        cframe.pack()
        bframe = Tk.Frame(body_frame)
        bframe.pack(fill=Tk.BOTH, expand=Tk.YES)
        tframe = Tk.Frame(bframe)
        tframe.pack(side=Tk.LEFT)
        pframe = Tk.Frame(bframe)
        pframe.pack(side=Tk.RIGHT)

        fig = plt.figure(figsize=(10,6))
        self.mpl_canvas = FigureCanvasTkAgg(fig, cframe)
        self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
        self.mpl_canvas_widget.pack(fill=Tk.BOTH, expand=1)

        self.ax = fig.add_subplot(111)
        self.build_figure(self.ax)

        fig.tight_layout()
        self.mpl_canvas.draw()

        self.toolbar = NavigationToolbar(self.mpl_canvas, tframe)
        self.toolbar.update()
        self.build_manipulation_parts(self.ax, pframe)

    def build_figure(self, ax):
        if self.outline_fig_frame is None:
            return

        self.outline_fig_frame.draw_outline_elution(ax)
        xmin, xmax = ax.get_xlim()
        ymin, ymax = ax.get_ylim()
        ax.set_xlim(xmin, xmax)
        ax.set_ylim(ymin, ymax)

        tx = xmin*0.5 + xmax*0.5
        ty = ymin*0.2 + ymax*0.8
        ax.text(tx, ty, "Drag to set the Window Range.", alpha=0.3, fontsize=30,
                ha="center", va="center")

    def build_manipulation_parts(self, ax, pframe):
        self.y = self.outline_fig_frame.exact_elution
        length = len(self.y)
        self.x = np.arange(length)

        if self.restrict is None:
            init_min = 0
            init_max = length-1
        else:
            init_min = self.restrict.start
            init_max = self.restrict.stop-1

        self.init_min = init_min
        self.init_max = init_max

        qmm_window_slice = get_setting('qmm_window_slice')

        if qmm_window_slice is None:
            min_ = init_min
            max_ = init_max
        else:
            min_ = qmm_window_slice.start
            max_ = qmm_window_slice.stop - 1
            self.draw_window(min_, max_)

        self.from_  = Tk.IntVar()
        self.from_.set(min_)
        self.to_  = Tk.IntVar()
        self.to_.set(max_)
        self.ivars  = [ self.from_, self.to_ ]

        k = -1
        for i, t in enumerate(["from ", "to "]):
            if i == 1:
                k += 1
                space = Tk.Label(pframe, width=1)
                space.grid(row=0, column=k)

            k += 1
            label = Tk.Label(pframe, text=t)
            label.grid(row=0, column=k)

            k += 1
            entry   = Tk.Spinbox( pframe, textvariable=self.ivars[i],
                        from_=init_min, to=init_max, increment=1, 
                        justify=Tk.CENTER, width=6 )
            entry.grid(row=0, column=k)

        self.spinbox_stop_trace = False
        for k, dvar in enumerate(self.ivars):
            dvar.trace('w', lambda *args, k_=k:self.spinbox_tracer(k_))

        self.span = SpanSelector(ax, self.onselect, 'horizontal', useblit=True,
                            props=dict(alpha=0.5))

        clear_button = Tk.Button(pframe, text="Clear", command=self.clear_window)
        clear_button.grid(row=0, column=5, padx=10)

    def draw_window(self, xmin, xmax):
        ax = self.ax
        ymin, ymax = ax.get_ylim()
        if self.window_patch is None:
            rect = Rectangle(
                    (xmin, ymin),   # (x,y)
                    xmax - xmin,    # width
                    ymax - ymin,    # height
                    facecolor   = 'cyan',
                    alpha       = 0.2,
                )
            self.window_patch = ax.add_patch(rect)
        else:
            rect = self.window_patch
            rect.set_xy((xmin, ymin))
            rect.set_width(xmax - xmin)
        self.mpl_canvas.draw()

    def clear_window(self):
        if self.window_patch is None:
            return

        self.window_patch.remove()
        self.window_patch = None
        self.spinbox_stop_trace = True
        self.from_.set(0)
        self.to_.set(len(self.x)-1)
        self.update()
        self.spinbox_stop_trace = True
        self.mpl_canvas.draw()

    def onselect(self, xmin, xmax):
        if xmax - xmin < 1e-6:
            return

        # print('onselect', (xmin, xmax))
        xmin = max(self.init_min, xmin)
        xmax = min(self.init_max, xmax)
        self.spinbox_stop_trace = True
        try:
            max_i = len(self.x)-1
            start = min(max_i, bisect_right(self.x, xmin))
            j = min(max_i, bisect_right(self.x, xmax))
            stop = j+1
            self.from_.set(self.x[start])
            self.to_.set(self.x[j])
            self.update()
            self.draw_window(xmin, xmax)
        except:
            if True:
                from molass_legacy.KekLib.ExceptionTracebacker import ExceptionTracebacker
                etb = ExceptionTracebacker()
                print(etb.last_lines())
        self.spinbox_stop_trace = False

    def spinbox_tracer(self, k):
        if self.spinbox_stop_trace:
            return

        values = []
        try:
            for ivar in self.ivars:
                values.append(ivar.get())
        except:
            return

        if values[0] > values[1]:
            print('values=', values)
            if k == 0:
                v = values[1]
            else:
                v = values[0]
            self.spinbox_stop_trace = True
            self.ivars[k].set(v)
            values[k] = v
            self.update()
            self.spinbox_stop_trace = False

        start = values[0]
        stop = values[1]
        self.draw_window(start, stop)

    def apply(self):
        if self.window_patch is None:
            qmm_window_slice = None
            set_setting('qmm_window_slice_uv', None)
        else:
            start = self.from_.get()
            stop = self.to_.get() + 1
            qmm_window_slice = slice(start, stop)
            self.set_uv_window(start, stop)
        set_setting('qmm_window_slice', qmm_window_slice)

    def set_uv_window(self, start, stop):
        if self.cs is None:
            qmm_window_slice_uv = None
        else:
            cs = self.cs
            i_list = []
            for j in [start, stop]:
                i_list.append(int(round(cs.slope*j + cs.intercept)))
            qmm_window_slice_uv = slice(*i_list)
        set_setting('qmm_window_slice_uv', qmm_window_slice_uv)



"""
    DualEvaluator.py

    Copyright (c) 2018-2024, SAXS Team, KEK-PF
"""
import copy
import numpy as np
from bisect import bisect_right
from molass_legacy.SerialAnalyzer.ElutionCurve import PEAK_INFO_FIND_RATIO
from ModelEvaluator import ModelEvaluator

class DualEvaluator(ModelEvaluator):
    def __init__(self, model, param_values, debug=False):
        self.model = model
        self.model_name = model.get_name()
        self.func = model.func
        if len(param_values) != 5:
            print("DualEvaluator: param_values=", param_values)
        assert len(param_values) == 5
        self.param_values = param_values
        self.sign = 1   # no need for negative sign?
        self.accepts_real_x = False     # for temporary fix under V1PreviewAdapter



"""

    ElutionModelUtils.py

    Copyright (c) 2017-2024, SAXS Team, KEK-PF

"""
from bisect import bisect_right
import numpy as np
# from scipy.interpolate import UnivariateSpline
from .InvertibleSpline import InvertibleSpline as UnivariateSpline

def compute_4moments(x, y):
    """
    note that this is diffrent from the one in Peaks.ElutionModels
    in the sense that this one returns W as well
    """
    W = np.sum(y)
    M1 = np.sum(x*y)/W              # raw moment
    M2 = np.sum(y*(x-M1)**2)/W      # central moment
    M3 = np.sum(y*(x-M1)**3)/W      # central moment
    return W, M1, M2, M3

def get_xies_from_height_ratio(alpha, x, y, max_y=None, debug=False):
    if max_y is None:
        max_y = np.max(y)

    alpha_y = max_y*alpha
    safe_range = y > alpha_y*0.5
    x_ = x[safe_range]
    y_ = y[safe_range]

    if debug and len(y_) == 0:
        import molass_legacy.KekLib.DebugPlot as plt
        with plt.Dp():
            fig, ax = plt.subplots()
            ax.plot(x, y)
            fig.tight_layout()
            plt.show()

    j = np.argmax(y_)       # ValueError: attempt to get argmax of an empty sequence

    if debug:
        import molass_legacy.KekLib.DebugPlot as plt
        # to plot UnivariateSpline below failure with pH6
        with plt.Dp():
            fig, ax = plt.subplots()
            ax.set_title("get_xies_from_height_ratio debug")
            ax.plot(x, y)
            for k, (f, t) in enumerate([(0, j-1), (j+1, len(x))]):
                xs = x_[f:t]
                ys = y_[f:t]
                ax.plot(xs, ys, ":")
            fig.tight_layout()
            plt.show()

    ret_xes = []
    for k, (f, t) in enumerate([(0, j-1), (j+1, len(x))]):
        xs = x_[f:t]
        ys = y_[f:t]
        try:
            if k == 0:
                spline = UnivariateSpline(ys, xs, s=0, ext=3)       # make the inverse function
            else:
                spline = UnivariateSpline(-ys, xs, s=0, ext=3)      # x must be strictly increasing
                alpha_y *= -1
            ret_xes.append(spline(alpha_y).item())    # numpy.ndarray.item: array to scalar
        except:
            # this case may be ignored
            from molass_legacy.KekLib.ExceptionTracebacker import log_exception
            log_exception(None, "get_xies_from_height_ratio: ", n=10)

    if debug:
        print("ret_xes=", ret_xes)
        spline = UnivariateSpline(x, y, s=0, ext=3)
        with plt.Dp():
            fig, ax = plt.subplots()
            ax.set_title("x_from_height_ratio_impl debug")
            ax.plot(x, y)
            ax.plot(x_, y_)
            for px in ret_xes:
                ax.plot(px, spline(px), "o", color="yellow")
            fig.tight_layout()
            plt.show()

    return ret_xes

def x_from_height_ratio_impl(func, ecurve, alpha, *params, needs_ymax=False, full_params=False,
                             x=None,
                             fx=None,   # currently used for Stochastic Models
                             debug=False):
    assert alpha > 0
    if x is None:
        x = ecurve.x
    if fx is None:
        fx = x
    if full_params:
        # for EDM, task: consider always doing it this way
        y = func(fx, *params)
    else:
        y = func(fx, 1, *params)

    max_y = np.max(y) if needs_ymax else 1       # EDM needs_ymax because its max(y) will not neccesarily be near 1

    return get_xies_from_height_ratio(alpha, x, y, max_y=max_y, debug=debug)



"""
    Optimizer.GlobalInspector.py

    Copyright (c) 2020,2023, SAXS Team, KEK-PF
"""
import logging
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.gridspec import GridSpec
from mpl_toolkits.mplot3d import Axes3D
from molass_legacy._MOLASS.SerialSettings import get_setting
from DataUtils import get_in_folder
from .OptimalElution import compute_optimal_elution
from MatrixData import simple_plot_3d
from RawData import get_row_index_impl
from molass_legacy.KekLib.OurTkinter import Tk, Dialog
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from molass_legacy.KekLib.OurMatplotlib import NavigationToolbar

BUG_FIX_WITH_REDRAW = False
TITLE_FONTSIZE = 16
MATRIX_SYM_FONTSIZE = 200

Q_LIST = [0.02, 0.05, 0.1]
MODELS = ["EMG", "EGH"]

FIG_SIZE = (21,12)
AXIS_NUM_H_DIVS = 5

class GlobalInspector:
    def __init__(self, D, E, ecurve, legacy_info=None, logger=None):
        self.logger = logging.getLogger(__name__) if logger is None else logger
        results = []
        norm_ratios = []
        for model_type in [0, 1]:
            M, rank, C_init, optC, recs = compute_optimal_elution(D, E, ecurve, model_type, legacy_info=legacy_info, logger=self.logger)
            P_init = M @ np.linalg.pinv(C_init)
            norm0 = np.linalg.norm(P_init@C_init - M)

            P = M @ np.linalg.pinv(optC)
            norm1 = np.linalg.norm(P@optC - M)
            norm_ratios.append(norm1/norm0)
            results.append((M, rank, optC, P, norm0, norm1, recs))
        self.results = np.array(results)
        self.norm_ratios = norm_ratios
        self.ecurve = ecurve

    def plot_results(self, fig, D, qvector, same_scale_residuals=True):
        in_folder = get_setting('in_folder')

        ecurve = self.ecurve
        results = self.results
        norm_ratios = self.norm_ratios

        folder_name = get_in_folder(in_folder=in_folder)
        fig.suptitle("Global Fitting Inspection for " + folder_name + "; Norm Reduction Ratios: EMG(%.3g), EGH(%.3g)" % tuple(norm_ratios), fontsize=20)
        nrows, ncols = 3, 3
        gs = GridSpec(nrows*AXIS_NUM_H_DIVS, ncols)
        axes = []
        axes_d = []
        for i in range(nrows):
            row = []
            row_d = []
            i_start = i*AXIS_NUM_H_DIVS
            for j in range(ncols):
                if i+j == 0:
                    projection = '3d'
                    i_stop = i_start + AXIS_NUM_H_DIVS
                else:
                    projection = None
                    i_stop = i_start + (AXIS_NUM_H_DIVS - 1)
                ax = fig.add_subplot(gs[i_start:i_stop,j], projection=projection)
                # ax.get_xaxis().set_visible(False)
                row.append(ax)
                if projection is None:
                    ax_d = fig.add_subplot(gs[i_stop,j])
                    ax_d.get_xaxis().set_visible(False)
                else:
                    ax_d = None
                row_d.append(ax_d)

            axes.append(row)
            axes_d.append(row_d)

        axes = np.array(axes)
        axes_d = np.array(axes_d)
        ax00 = axes[0,0]
        ax01 = axes[0,1]
        ax02 = axes[0,2]

        sub_fontsize = 16
        std_Q = Q_LIST[0]

        x = ecurve.x
        y = ecurve.y

        ax00.set_title("3D View", y=1.07, fontsize=sub_fontsize)
        simple_plot_3d(ax00, D, x=qvector)

        residual_ylims = []
        residual_indeces = []
        for j, (name, ax, recs) in enumerate(zip(MODELS, axes[0,1:], results[:,-1])):
            ax.set_title("%s Initial Decomposition at %.2g" % (name, std_Q), y=0.98, fontsize=sub_fontsize)
            ax.plot(x, y, color='orange', alpha=0.5, linewidth=3, label='data')
            ty = np.zeros(len(y))
            for k, rec in enumerate(recs):
                f = rec.evaluator
                y_ = f(x)
                ax.plot(x, y_, ':', linewidth=3, label='element %d' % (k+1))
                ty += y_
            ax.plot(x, ty, ':', color='red', linewidth=3, label='model total')
            ax.legend()
            index_d = (0,j+1)
            ax_d = axes_d[index_d]
            ax_d.plot(x, ty-y, color='gray', label='residual')
            ax_d.legend(bbox_to_anchor=(1, 0.9), loc='lower right')
            residual_indeces.append(index_d)
            residual_ylims.append(ax_d.get_ylim())

        average_width = get_setting('num_points_intensity')
        hw = average_width//2

        for i, (M, rank, optC, P, norm0, norm1, recs) in enumerate(results):
            name = MODELS[i]
            for j, qv in enumerate(Q_LIST):
                ax = axes[i+1,j]
                ax.set_title("%s Result Decomposition at %.2g" % (name, qv), y=0.98, fontsize=sub_fontsize)
                k = get_row_index_impl(qvector, qv)
                ey = np.average(D[k-hw:k+hw+1,:], axis=0)
                ax.plot(x, ey, color='orange', alpha=0.5, linewidth=3, label='data')

                ty = np.zeros(len(y))
                if True:
                    for m, c in enumerate(optC):
                        scale = P[k,m]
                        y_ = scale*c
                        ax.plot(x, y_, ':', linewidth=3, label='element %d' % (m+1))
                        ty += y_
                else:
                    for m, rec in enumerate(recs):
                        scale = P[k,m]
                        f = rec.evaluator
                        y_ = scale*f(x)
                        ax.plot(x, y_, ':', linewidth=3, label='element %d' % (m+1))
                        ty += y_
                ax.plot(x, ty, ':', color='red', linewidth=3, label='model total')
                ax.legend()
                index_d = (i+1,j)
                ax_d = axes_d[index_d]
                ax_d.plot(x, ty-ey, color='gray', label='residual')
                ax_d.legend(bbox_to_anchor=(1, 0.9), loc='lower right')
                residual_indeces.append(index_d)
                residual_ylims.append(ax_d.get_ylim())

        if same_scale_residuals:
            ylims = np.array(residual_ylims)
            ymin = np.min(ylims[:,0])
            ymax = np.max(ylims[:,1])
            w = -0.1
            ymin_ = ymin*(1-w) + ymax*w
            w = 1.1
            ymax_ = ymin*(1-w) + ymax*w
            for index in residual_indeces:
                axes_d[index].set_ylim(ymin_, ymax_)

        fig.subplots_adjust(top=0.93, bottom=0.02, left=0.05, right=0.97, hspace=1.1)

class GlobalInspectorDialog(Dialog):
    def __init__(self, parent, dialog, sd):
        D, E, qvector, ecurve = sd.get_xr_data_separate_ly()
        legacy_info = [sd, dialog.mapper, dialog.corbase_info]
        self.inspector = GlobalInspector(D, E, ecurve, legacy_info)
        self.D = D
        self.qvector = qvector

        Dialog.__init__(self, parent, "Global Inspector", visible=False)

    def show(self):
        self._show()

    def body(self, body_frame):
        cframe = Tk.Frame(body_frame)
        cframe.pack()
        bframe = Tk.Frame(body_frame)
        bframe.pack(fill=Tk.X)
        tframe = Tk.Frame(bframe)
        tframe.pack(side=Tk.LEFT)
        btn_frame = Tk.Frame(bframe)
        btn_frame.pack(side=Tk.RIGHT, padx=40)

        self.fig = fig = plt.figure(figsize=FIG_SIZE)
        self.mpl_canvas = FigureCanvasTkAgg(fig, cframe)
        self.mpl_canvas_widget = self.mpl_canvas.get_tk_widget()
        self.mpl_canvas_widget.pack( fill=Tk.BOTH, expand=1 )
        self.toolbar = NavigationToolbar(self.mpl_canvas, tframe)
        self.toolbar.update()

        self.inspector.plot_results(fig, self.D, self.qvector)
        self.mpl_canvas.draw()

        self.add_buttons(btn_frame)

    def buttonbox(self):
        pass

    def add_buttons(self, frame):
        self.same_scale = Tk.IntVar()
        self.same_scale.set(1)
        w = Tk.Checkbutton(frame, text="Same scale for residuals", variable=self.same_scale)
        w.pack(side=Tk.LEFT, padx=10)
        w = Tk.Button(frame, text="Close", command=self.cancel)
        w.pack(side=Tk.LEFT, padx=10)
        self.same_scale.trace("w", self.same_scale_tracer)

    def same_scale_tracer(self, *args):
        same_scale_residuals = self.same_scale.get() == 1
        fig = self.fig
        fig.clf()
        self.inspector.plot_results(fig, self.D, self.qvector, same_scale_residuals)
        self.mpl_canvas.draw()



"""
    SplashMessage.py

    Copyright (c) 2017-2025, SAXS Team, KEK-PF
"""

import tkinter as Tk
import tkinter.ttk as ttk
from molass_legacy.KekLib.TkUtils import adjusted_geometry
from molass_legacy.KekLib.MultiMonitor import get_selected_monitor
from molass_legacy.KekLib.SplashScreen import SplashScreen

splash_message1 = ( "MOLASS is loading...\n"
                    "Please be patient."
                    )

splash_geometry = '600x380+200+200'

class AppSplashScreen:
    def __init__( self, root, maxval, value ):
        self.root = root
        sp = SplashScreen(root)
        # sp.config(bg="#3366ff")

        pad = Tk.Frame( sp, height=50 )
        pad.pack()

        m = Tk.Label( sp, text=splash_message1, justify=Tk.CENTER, font=("calibri", 30) )
        m.pack()

        self.bar = bar = ttk.Progressbar(sp, orient='horizontal', length=400, mode='determinate')
        bar.pack( pady=40 )
        bar["value"]    = value
        bar["maximum"]  = maxval

        monitor = get_selected_monitor()
        if monitor.width >= 2560:
            # tempoary fix for high-DPI
            # splash_geometry_ = '900x570+300+300'
            splash_geometry_ = splash_geometry
        else:
            splash_geometry_ = splash_geometry
        geometry_ = adjusted_geometry(splash_geometry_)
        root.geometry( geometry_ )
        root.update()

    def update( self ):
        self.bar["value"] += 1
        self.root.update()



# coding: utf-8
"""
    CormapMaker.py

    Copyright (c) 2021-2022, SAXS Team, KEK-PF
"""

import numpy as np
from matplotlib import colors
from MplUtils import BoundaryNorm

class CormapMaker:
    def __init__(self, M, qv, seqno, seqname, transpose=False, from_datcmp=False):
        self.transpose = transpose
        self.from_datcmp = from_datcmp
        if from_datcmp:
            self.cormap = M
        else:
            self.cormap = self.compute_cormap(M.T if transpose else M)
        self.qv = qv
        self.seqno = seqno
        self.seqname = seqname

    def draw(self, ax, cbar_ax):
        if self.transpose:
            label = self.seqname
            f, t = self.seqno[[0,-1]]
        else:
            label = r"$q(\AA^{-1})$"
            f, t = self.qv[[0,-1]]
        ax.set_xlabel(label)
        ax.set_ylabel(label)
        if self.from_datcmp:
            """
            https://stackoverflow.com/questions/9707676/defining-a-discrete-colormap-for-imshow-in-matplotlib
            """
            cmap = colors.ListedColormap(["green", "palegreen", "tomato"])
            bounds=[-1, -0.1, 0.1, 1]
            norm = BoundaryNorm(bounds, cmap.N)
        else:
            cmap = None
            norm = None
        im = ax.imshow(self.cormap, extent=[f, t, t, f], cmap=cmap, norm=norm)
        fig = ax.get_figure()
        if self.from_datcmp:
            colorbar = fig.colorbar(im, ax=cbar_ax, ticks=[-0.5, 0, 0.5])
            colorbar.ax.set_yticklabels(["good", "ok", "bad"])  # vertically oriented colorbar
        else:
            colorbar = fig.colorbar(im, ax=cbar_ax)
        return colorbar

    def compute_cormap(self, M):
        cormap = np.corrcoef(M)
        return cormap



# coding: utf-8
"""

    extrapolation-demo-2.py

    Copyright (c) 2017, Masatsuyo Takahashi, KEK-PF

"""
import sys
import os
import numpy    as np
from mpl_toolkits.mplot3d   import Axes3D
import matplotlib.pyplot    as plt
# import statsmodels.api      as sm
from Intensity              import Intensity

this_dir = os.path.dirname( os.path.abspath( __file__ ) )
sys.path.append( this_dir + '/../../lib' )
import molass_legacy.KekLib
import OurStatsModels       as sm
from molass_legacy.KekLib.NumpyUtils             import np_savetxt

SLICE_NUM_POINTS = 400
LOG_PLOT_A = True

x = np.linspace( 0.005, 0.8, 800 )
C = 1e-2
cvector_all = 0.1 * np.exp( - C * ( np.arange( 50 ) - 25 )**2  )
indeces = np.arange( 0, 25 )

e = np.ones( len(x) ) * 1e-3

intensity_list = []
d_error_list = []
for i in indeces:
    c = cvector_all[i]
    intensity = Intensity( c, 29 )
    ii = intensity( x )
    intensity_list.append( ii )
    d_error_list.append( e )

intensity_array = np.array( intensity_list )
d_error_array = np.array( d_error_list )
print( 'intensity_array.shape=', intensity_array.shape )
print( 'd_error_array.shape=', d_error_array.shape )

x_slice = slice( 0, SLICE_NUM_POINTS )
x_ = x[x_slice]

x_i = 50
x_f = x_[x_i]

fig = plt.figure( figsize=(18, 9) )
ax1 = fig.add_subplot( 231 )
ax2 = fig.add_subplot( 232, projection='3d' )
ax3 = fig.add_subplot( 233 )
ax4 = fig.add_subplot( 234, projection='3d' )
ax5 = fig.add_subplot( 235 )
ax6 = fig.add_subplot( 236 )

ax1.set_title( 'Concentration Curve' )
ax1.set_xlabel( 'Seq No.' )
ax1.set_ylabel( 'Concentration' )

ax2.set_title( 'Intensity Curves on Q[0:200] and the plane at Q[50]=%.3g' % x_f )
ax2.set_xlabel( 'Q' )
ax2.set_ylabel( 'Seq No.' )
ax2.set_zlabel( 'Intensity' )

ax3.set_title( 'Intensity Variation with errorbar at Q[50]=%.3g' % x_f )
ax3.set_xlabel( 'Seq No.' )
ax3.set_ylabel( 'Concentration' )

ax4.set_title( 'Mutivariate Regression at Q[50]=%.3g' % x_f )
ax4.set_xlabel( 'C' )
ax4.set_ylabel( 'C²' )
ax4.set_zlabel( 'Intensity' )

ax5.set_title( 'Extrapolated A(q)' )
ax5.set_xlabel( 'Q' )
ylabel_ = 'ln(Intensity / C)' if LOG_PLOT_A else 'Intensity / C'
ax5.set_ylabel( ylabel_ )

ax6.set_title( 'Extrapolated B(q)' )
ax6.set_xlabel( 'Q' )
ax6.set_ylabel( 'Intensity / C²' )


cvector_ = cvector_all[indeces]
ax1.plot( indeces, cvector_ )

for i in indeces:
    c = cvector_all[i]
    # y_ = np.ones( len(x_) ) * c
    y_ = np.ones( len(x_) ) * i
    z_ = intensity_array[i,x_slice]
    ax2.plot( x_, y_, z_ )

ymin, ymax = ax2.get_ylim()
ax2.set_ylim( ymin, ymax )
zmin, zmax = ax2.get_zlim()
ax2.set_zlim( zmin, zmax )
ax2.plot( np.ones(5)*x_f,
            [ymin, ymax, ymax, ymin, ymin ],
            [zmin, zmin, zmax, zmax, zmin ],
            ':', color='black' )
ax2.plot( np.ones(len(cvector_))*x_f, indeces, intensity_array[indeces,x_i], ':', color='red' )

y_f = intensity_array[indeces,x_i]
y_e = d_error_array[indeces,x_i]
ax3.plot( indeces, y_f, ':', color='red' )
# ax3.errorbar( indeces, y_f, yerr=y_e, color='blue', fmt='none' )

x1_ = cvector_
x2_ = cvector_**2
X = np.array( [ x1_, x2_] ).T

param_list = []
error_list = []
for i in range(0,SLICE_NUM_POINTS):
    z_f = intensity_array[indeces,i]
    e2  = d_error_array[indeces,i]**2
    w   = 1/e2
    model   = sm.WLS(z_f, X, weights=w)
    result  = model.fit()
    param_list.append( result.params )
    # error_list.append( np.diag( result.cov_params() ) )
    error = np.sqrt( np.dot( model.XtWX_inv_XtW**2, e2 ) )
    error_list.append( error )

params = param_list[x_i]

z_f = intensity_array[indeces,x_i]
z_r = np.dot( X, params )

ax4.scatter( x1_, x2_, z_f )
ax4.scatter( x1_, x2_, z_r, color='orange' )

xmin, xmax = ax4.get_xlim()
ymin, ymax = ax4.get_ylim()
ax4.set_xlim( xmin, xmax )
ax4.set_ylim( ymin, ymax )

xx, yy = np.meshgrid( np.linspace(xmin, xmax, 10), np.linspace(ymin, ymax, 10) )
zz = xx * params[0] + yy * params[1]
ax4.plot_surface(xx, yy, zz, alpha=0.1 )

params_array = np.array( param_list )
error_array = np.array( error_list )
# np_savetxt( 'params_array.csv', params_array )
# np_savetxt( 'error_array.csv', error_array )

A = params_array[:,0]
B = params_array[:,1]
A_ = np.log(A) if LOG_PLOT_A else A
ax5.plot( x_, A_, color='red', label='Kek A(q)' )
ax6.plot( x_, B, color='red', label='Kek B(q)' )
ax5.errorbar( x_f, A[x_i], yerr=error_array[x_i,0], color='blue', fmt='none' )
ax6.errorbar( x_f, B[x_i], yerr=error_array[x_i,1], color='blue', fmt='none' )

c = 1
intensity = Intensity( c, 29 )
At = intensity.term1( x_ )
Bt = -At*intensity.term2( x_ )

ax5.plot( x_, np.log(At), ':', color='cyan', label='True A(q)' )
ax6.plot( x_, Bt, ':', color='cyan', label='True B(q)' )

# np_savetxt( 'A.csv', A )
# np_savetxt( 'B.csv', B )

for ax in [ax5, ax6]:
    ymin, ymax = ax.get_ylim()
    ax.set_ylim( ymin, ymax )
    ax.plot( [ x_f, x_f ], [ymin, ymax], ':', color='black' )

for ax in [ax5, ax6]:
    ax.legend()

plt.tight_layout()
plt.show()



"""

    DatgnomExcelFormatter.py

    SummaryBook.py

    Copyright (c) 2018-2025, SAXS Team, KEK-PF

"""
from molass_legacy.KekLib.ExcelCOM import xlCategory, xlValue, xlLow, msoThemeColorText1
from molass_legacy.KekLib.ExcelLogChart import ExcelLogChart
from .DefaultFont import set_default_bookfont_com

class DatgnomResultArgs:
    def __init__(self, title, book_file):
        self.title = title
        self.book_file = book_file

def add_datgnom_format_setting(excel_client_, args, logger):
    # logger is not yet used

    ws = excel_client_.openWorksheet( args.book_file + '(' + args.title  + ')' )
    set_default_bookfont_com(ws.workbook)

    chart = ws.get_chart( 0 )
    # plot area frame lines
    line = chart.PlotArea.Format.Line
    line.Visible    = True
    line.ForeColor.ObjectThemeColor = msoThemeColorText1

    # grid lines
    chart.Axes(xlValue).MajorGridlines.Format.Line.Visible = False
    chart.Axes(xlCategory).MajorGridlines.Format.Line.Visible = False

    # text font size
    chart.ChartArea.Format.TextFrame2.TextRange.Font.Size = 16

    y_axis = chart.Axes(xlValue)
    # y_axis.TickLabels.NumberFormatLocal = "#,##0.000_)"
    # y_axis.TickLabels.NumberFormatLocal = "0.0E+00"
    y_axis.MajorUnit    = 10
    chart.Axes(xlCategory).TickLabelPosition = xlLow

    log_chart = ExcelLogChart( chart, ws.worksheet )
    log_chart.change_y_labels( superscripts=True )
    L, T, W, H = log_chart.get_plotarea()
    # print( 'L, T, W, H=', L, T, W, H )
    log_chart.change_plotarea( left=L+30, width=W-50 )
    log_chart.change_yaxis_title_area( left=25 )

    chart = ws.get_chart( 1 )
    # plot area frame lines
    line = chart.PlotArea.Format.Line
    line.Visible    = True
    line.ForeColor.ObjectThemeColor = msoThemeColorText1

    # grid lines
    chart.Axes(xlValue).MajorGridlines.Format.Line.Visible = False
    chart.Axes(xlCategory).MajorGridlines.Format.Line.Visible = False

    # text font size
    chart.ChartArea.Format.TextFrame2.TextRange.Font.Size = 16

    y_axis = chart.Axes(xlValue)
    # y_axis.TickLabels.NumberFormatLocal = "#,##0.000_)"
    y_axis.TickLabels.NumberFormatLocal = "0.0E+00"

    ws.zoom( 70 )
    # scroll_column=16
    ws.freeze_panes( "3:3", scroll_column=None )

    ws.workbook.Save()
    ws.workbook.Close()



# coding: utf-8
"""
    QuantileRegressionKeras.py

    Copyright (c) 2019, SAXS Team, KEK-PF
"""
import numpy as np
from keras.models import Sequential
from keras.layers import Dense, Activation
import keras.backend as K

def tilted_loss(q,y,f):
    e = (y-f)
    return K.mean(K.maximum(q*e, (q-1)*e), axis=-1)

class Model:
    def __init__(self):
        self.model = model = Sequential()
        model.add(Dense(units=10, input_dim=1,activation='relu'))
        model.add(Dense(units=10, input_dim=1,activation='relu'))
        model.add(Dense(1))
        # model.compile(loss='mae', optimizer='adadelta')
        model.compile(loss=lambda y,f: tilted_loss(0.5,y,f), optimizer='adadelta')

    def fit(self, x, y):
        self.model.fit(x, y, epochs=1000, batch_size=32, verbose=0)
        self.model.evaluate(x, y)

    def predict(self, x):
        return self.model.predict(x)



# coding: utf-8
"""
    DecompResetDialog.py

    Copyright (c) 2019, SAXS Team, KEK-PF
"""
import re
import os
import logging
from molass_legacy.KekLib.OurTkinter import Tk, Dialog
from molass_legacy.KekLib.TkSupplements import set_icon
from molass_legacy.KekLib.TkCustomWidgets import FileEntry
from molass_legacy.KekLib.ReadOnlyText import ReadOnlyText
from DecompEditorFrame import DECOMP_EDITOR_LOG_HEADER, DECOMP_EDITOR_LOG_TRAILER
from molass_legacy._MOLASS.SerialSettings import get_setting

in_folder_re    = re.compile(r'^.+started loading (\S+)')
log_line_re     = re.compile(r'^.+DecompEditorDialog,')
model_name_re   = re.compile(r'^.+model_name=(\w+)')

params_re   = re.compile(r'\[\d+\] h=(\S+), mu=(\S+), sigma=(\S+), tau=(\S+), a=(\S+); with tau constraints \((\S+), (\S+)\) and uv/xray h ratio (\S+)')
range_re    = re.compile(r'\[\d+\] (\[.*\]) with element curves (\[.*\])')

def extract_params_from_log_lines(lines):
    param_recs = []
    range_recs = []
    for line in lines:
        m = params_re.search(line)
        if m:
            param_recs.append([ float(m.group(i)) for i in range(1,9) ])
            continue
        m = range_re.search(line)
        if m:
            range_recs.append([eval(m.group(i)) for i in range(1,3) ])
            continue
    return param_recs, range_recs

class DecompResetDialog( Dialog ):
    def __init__(self, parent):
        self.logger  = logging.getLogger( __name__ )
        self.parent = parent
        self.in_folder = None
        self.canceled = False
        self.applied = False
        Dialog.__init__( self, parent, "DecompResetDialog", visible=False )

    def show(self):
        self._show()

    def body(self, body_frame):
        filepath_frame = Tk.Frame(body_frame)
        filepath_frame.pack()

        file_label = Tk.Label(filepath_frame, text="Log File Path: ")
        file_label.pack(side=Tk.LEFT)
        self.logfile_path = Tk.StringVar()
        file_entry = FileEntry(filepath_frame, textvariable=self.logfile_path, width=60, on_entry_cb=self.on_entry )
        file_entry.pack(side=Tk.LEFT)

        current_log_btn = Tk.Button(filepath_frame, text="Current Log File", command=self.set_current_logfile)
        current_log_btn.pack(side=Tk.LEFT, padx=10)

        space = Tk.Frame(body_frame)
        space.pack(pady=5)

        text_label = Tk.Label(body_frame, text="The following info has been retrieved from the log file.")
        text_label.pack()
        self.ro_text = ReadOnlyText(body_frame, width=140, height=16)
        self.ro_text.pack(padx=10)

        space = Tk.Frame(body_frame)
        space.pack(pady=5)

        guide_label = Tk.Label(body_frame, text='If you would like to reset with this info, please press "Reset", otherwise "Cancel".')
        guide_label.pack()

        soon_label = Tk.Label(body_frame, text='"Reset", disabled in this version, will soon be available.', fg='orange')
        soon_label.pack()

    def buttonbox(self):

        box = Tk.Frame(self)
        box.pack()

        w = Tk.Button(box, text="Reset", width=10, command=self.reset, default=Tk.ACTIVE, state=Tk.DISABLED)
        w.pack(side=Tk.LEFT, padx=5, pady=5)
        w = Tk.Button(box, text="Cancel", width=10, command=self.cancel)
        w.pack(side=Tk.LEFT, padx=5, pady=5)

        self.bind("<Return>", self.ok)
        self.bind("<Escape>", self.cancel)

    def on_entry(self):
        file_path = self.logfile_path.get()
        print('on_entry', file_path)
        reading_target_log = False
        model_name = None
        fh = open(file_path)
        for line in fh.readlines():
            if line.find('started loading') > 0:
                m = in_folder_re.search(line)
                self.in_folder = m.group(1)

            if line.find(DECOMP_EDITOR_LOG_HEADER) > 0:
                reading_target_log = True
                model_name = None
                info_lines = []
                continue
            if line.find(DECOMP_EDITOR_LOG_TRAILER) > 0:
                reading_target_log = False

            if reading_target_log:
                cut_line = re.sub(log_line_re, '', line)
                m = model_name_re.search(cut_line)
                if m:
                    model_name = m.group(1)
                info_lines.append(cut_line)
        fh.close()

        self.ro_text.delete('1.0', Tk.END)

        in_folder_info = 'unretrieved' if self.in_folder is None else self.in_folder
        self.ro_text.insert(Tk.END, 'in_folder=' + in_folder_info + '\n\n')
        try:
            for cut_line in info_lines:
                self.ro_text.insert(Tk.END, cut_line)
        except:
            pass

    def set_current_logfile(self):
        import glob
        analysis_folder = get_setting('analysis_folder')
        try:
            logfiles = glob.glob(analysis_folder + '/*.log')
            self.logfile_path.set(logfiles[0].replace('\\', '/'))
            self.on_entry()
        except:
            from molass_legacy.KekLib.ExceptionTracebacker import ExceptionTracebacker
            etb = ExceptionTracebacker()
            import OurMessageBox as MessageBox
            MessageBox.showerror("Error", "Couldn't find the current log file.\n---\n" + str(etb), parent=self)

    def reset(self):
        pass



# coding: utf-8
"""
    Rgg.MostSimultaneous.py

    Copyright (c) 2021, SAXS Team, KEK-PF
"""
import os
import logging
from time import time
import numpy as np
from scipy.optimize import minimize, basinhopping
from scipy.stats import linregress
import molass_legacy.KekLib.DebugPlot as plt
from molass_legacy.Peaks.ElutionModels import egh_pdf

MAPPING_PENALTY_SCALE = 1e2
VANISH_PENALTY_SCALE = 1e3
VANISH_BOUND = 0.05
TAU_LIMIT_RATIO = 0.5   # relative to sigma
ASYMMETRY_PENALTY_SCALE = 1e2
SIGMA_LIMIT_RAIO = 1.0
SIGMA_PENALTY_SCALE = 1e2
MIN_LOG10_VALUE = 1e-10
EVAL_PARAMS_DEBUG = False
USE_DLIB = False
if USE_DLIB:
    import dlib
    from molass_legacy.KekLib.BasicUtils import Struct

class SimultaneousSolver:
    def __init__(self, n_components, xr_curve, rg_curve, uv_curve, min_ratio=0.02):
        self.logger = logging.getLogger(__name__)
        self.n_components = n_components
        self.xr_curve = xr_curve
        self.rg_curve = rg_curve
        self.uv_curve = uv_curve
        self.min_ratio = min_ratio
        self.counter = -1
        if rg_curve is None:
            self.slices = None
        else:
            self.slices = rg_curve.get_valid_slices()

    def make_random_proportions(self, seed=None):
        if seed is None:
            if self.given_seeds is None:
                seed = np.random.randint(1000, 9999)
            else:
                seed = self.given_seeds[0]
        self.peaks_seed = seed
        np.random.seed(seed)
        proportions = sorted(np.random.uniform(0, 1, self.n_components))
        return proportions

    def guess_init_curve_params(self, init_proportions, curve, save_sigma=False, debug=False):
        mean, v = curve.compute_moments()
        s = np.sqrt(v)
        if save_sigma:
            self.xr_sigma = s

        x = curve.x
        y = curve.y
        max_y = curve.max_y
        xmin = max(x[0], mean - 5*s)
        xmax = min(x[-1], mean + 5*s)
        peak_topx = []
        for k, w in enumerate(init_proportions):
            x_ = xmin * (1-w) + xmax*w
            if debug:
                print([k], x_)
            peak_topx.append(x_)

        peak_topx = np.array(peak_topx)
        peak_topy= np.max([np.zeros(len(peak_topx)), curve.spline(peak_topx)], axis=0)
        w_ = np.power(peak_topy, 1/self.n_components)   # make them less distinctive
        weights = w_/np.sum(w_)

        def obj_func(heights):
            ty = np.zeros(len(y))
            for h, m, w in zip(heights, peak_topx, weights):
                ty += h * egh_pdf(x, m, s*w, 0)
            return np.sum((ty - y)**2)

        bounds = [(0, None)] * self.n_components
        try:
            result = minimize(obj_func, peak_topy, bounds=bounds)
        except Exception as exc:
            from molass_legacy.KekLib.ExceptionTracebacker import log_exception
            log_exception(self.logger, "minimize")
            print(mean, s, peak_topx, peak_topy, weights)
            raise exc

        if debug:
            plt.push()
            fig, ax = plt.subplots()
            ax.plot(x, y)
            ty = np.zeros(len(y))
            for h, m, w in zip(result.x, peak_topx, weights):
                y_ = h * egh_pdf(x, m, s*w, 0)
                ax.plot(x, y_, ':')
                ty += y_
            ax.plot(x, ty, ':', color='red')
            fig.tight_layout()
            plt.show()
            plt.pop()

        return result.x, peak_topx, s*weights

    def guess_init_rgs(self, init_proportions, rg_curve):
        x_, y_, rg_ = rg_curve.get_valid_curves()
        mean = np.mean(rg_)
        std = np.std(rg_)
        return np.linspace(mean+std, mean-std, self.n_components)

    def guess_init_mapping(self):
        xr_topx = self.xr_curve.primary_peak_x
        uv_topx = self.uv_curve.primary_peak_x
        x = np.array([self.xr_curve.x[0], xr_topx, self.xr_curve.x[-1]])
        y = np.array([self.uv_curve.x[0], uv_topx, self.uv_curve.x[-1]])
        slope, intercept, r_value, p_value, std_err = linregress(x, y)
        return np.array([slope, intercept])

    def select_hopeful_init_params(self, debug=False):

        min_score = None
        min_init_params = None
        min_separate_params = None
        min_seed = None
        x = self.xr_curve.x
        init_mappable_range = (x[0], x[-1])

        k = 0
        while k < 10:
            init_proportions = self.make_random_proportions()
            try:
                init_xr_params = self.guess_init_curve_params(init_proportions, self.xr_curve, save_sigma=True)
                init_rgs = self.guess_init_rgs(init_proportions, self.rg_curve)
                init_uv_params = self.guess_init_curve_params(init_proportions, self.uv_curve)
                heights, peak_topx, sigmas = init_xr_params
                xr_params = []
                for h, m, s in zip(heights, peak_topx, sigmas):
                    xr_params.append((h, m, s, 0))
                init_mapping = self.guess_init_mapping()
                init_params = np.concatenate([np.array(xr_params).flatten(), init_rgs, init_mapping, init_uv_params[0], init_mappable_range])
                self.prepare_for_optimization(init_params)
                score = self.objective_func(init_params, debug=debug)
                print([k], "score=", score)
            except:
                continue
            if np.isnan(score):
                continue
            k += 1
            if min_score is None or score < min_score:
                min_score = score
                min_init_params = init_params
                min_separate_params = [xr_params, init_rgs, init_mapping, init_uv_params]
                min_seed = self.peaks_seed
                if self.given_seeds is not None:
                    break

        self.peaks_seed = min_seed
        return min_init_params, min_separate_params

    def solve(self, init_proportions=None, init_xr_params=None, init_rgs=None, init_mapping=None, init_uv_params=None,
                seeds=None, min_overlap=True, hopeful_only=True, debug=False):
        self.counter += 1
        self.given_seeds = seeds
        self.min_overlap = min_overlap
        self.debug_fv = False
        x = self.xr_curve.x
        init_mappable_range = (x[0], x[-1])

        if hopeful_only:
            init_params, separate_params = self.select_hopeful_init_params(debug=EVAL_PARAMS_DEBUG)
        else:
            if init_proportions is None:
                init_proportions = self.make_random_proportions()
            print("init_proportions=", init_proportions)

            if init_xr_params is None:
                init_xr_params = self.guess_init_curve_params(init_proportions, self.xr_curve, debug=debug)
            print("init_xr_params=", init_xr_params)

            if init_rgs is None:
                init_rgs = self.guess_init_rgs(init_proportions, self.rg_curve)
            print("init_rgs=", init_rgs)

            if init_uv_params is None:
                init_uv_params = self.guess_init_curve_params(init_proportions, self.uv_curve, debug=debug)
            print("init_uv_params=", init_uv_params)

            heights, peak_topx, sigmas = init_xr_params
            xr_params = []
            for h, m, s in zip(heights, peak_topx, sigmas):
                xr_params.append((h, m, s, 0))

            init_mapping = self.guess_init_mapping()
            separate_params = [xr_params, init_rgs, init_mapping, init_uv_params]
            init_params = np.concatenate([np.array(xr_params).flatten(), init_rgs, init_mapping, init_uv_params[0], init_mappable_range])

        self.prepare_for_optimization(init_params)

        if debug:
            self.objective_func(init_params, plot=True)

        if self.given_seeds is None:
            seed = np.random.randint(1000, 9999)
        else:
            seed = self.given_seeds[1]

        self.bh_seed = seed

        self.debug_fv = False       # self.debug_fv might have been set to True in other debug plots

        if USE_DLIB:
            lb, ub = self.make_bounds(separate_params)
            x, y = dlib.find_min_global(self.objective_func_for_dlib, 
                           lb,      # Lower bound constraints on x0 and x1 respectively
                           ub,      # Upper bound constraints on x0 and x1 respectively
                           10)          # T
            result = Struct(x=x, fun=y)
        else:
            result = basinhopping(self.objective_func, init_params, seed=seed)

        if debug:
            self.objective_func(result.x, plot=True)

        return result

    def get_used_seeds(self):
        return self.peaks_seed, self.bh_seed

    def split_params(self, p):
        n = self.n_components
        sep = 4*n
        xr_params = p[0:sep].reshape((n,4))
        rgs = p[sep:sep+n]
        sep = sep+n
        mapping = p[sep:sep+2]
        sep = sep+2
        uv_params = p[sep:sep+n]
        mappable_range = p[sep+n:]
        return xr_params, rgs, mapping, uv_params, mappable_range

    def prepare_for_optimization(self, init_params):
        self.xm, self.ym, self.rg = self.rg_curve.get_valid_curves()
        self.mask = self.rg_curve.get_mask()

        xr_params, rgs, mapping, uv_params, mappable_range = self.split_params(init_params)
        self.init_mapping = mapping

    def make_bounds(self, separate_params):
        xr_params, init_rgs, init_mapping, init_uv_params = separate_params
        lb = []
        ub = []
        assert False

        return lb, ub

    def objective_func_for_dlib(self, *args):
        return self.objective_func(np.array(args))

    def objective_func(self, p, debug=False, fig_info=None):
        debug_ = debug or self.debug_fv

        x = self.xr_curve.x
        y = self.xr_curve.y

        xm = self.xm
        rg = self.rg

        xr_params, rg_params, (a, b), uv_params, (c, d) = self.split_params(p)

        map_slice = slice(max(0,int(np.floor(c))), min(len(x),int(np.ceil(d))))

        uv_ty = np.zeros(len(y))
        ty = np.zeros(len(y))
        k = 0
        cy_list = []
        if debug_:
            uv_cy_list = []
        uv_x = a*x+b
        asymmetry_penalty = 0
        sigma_penalty = 0
        for h, m, s, t in xr_params:
            uv_cy = uv_params[k] * egh_pdf(uv_x, a*m+b, a*s, a*t)
            uv_ty += uv_cy
            cy = h * egh_pdf(x, m, s, t)
            cy_list.append(cy)
            if debug_:
                uv_cy_list.append(uv_cy)
            ty += cy
            asymmetry_penalty += max(0, t/s - TAU_LIMIT_RATIO)**2
            sigma_penalty += max(0, s/self.xr_sigma - SIGMA_LIMIT_RAIO)**2
            k += 1

        t_rg = np.zeros(len(rg))
        tym = ty[self.mask]
        for r, cy in zip(rg_params, cy_list):
            t_rg += r * cy[self.mask]/tym

        xr_scale = np.sum(ty**2)
        rg_scale = np.sum(t_rg**2)
        uv_scale = np.sum(uv_ty**2)

        uv_y = self.uv_curve.spline(uv_x)
        resid_xr = np.log10(np.sum((ty - y)[map_slice]**2)/xr_scale)
        resid_rg = np.log10(np.sum((t_rg -rg)**2)/rg_scale)
        resid_uv = np.log10(np.sum((uv_ty - uv_y)[map_slice]**2)/uv_scale)
        ratio = a/self.init_mapping[0]
        mapping_penalty = MAPPING_PENALTY_SCALE * ( min(0, ratio - 0.7)**2 + max(0, ratio - 1.4)**2 )
        xr_h_total = np.sum(xr_params[:,0])
        uv_h_total = np.sum(uv_params)
        vanishing_penalty = VANISH_PENALTY_SCALE * ( (min(VANISH_BOUND, np.min(xr_params[:,0])/xr_h_total) - VANISH_BOUND)**2 + (min(VANISH_BOUND, np.min(uv_params)/uv_h_total)- VANISH_BOUND)**2)
        asymmetry_penalty *= ASYMMETRY_PENALTY_SCALE
        sigma_penalty *= SIGMA_PENALTY_SCALE
        disorder_penalty = np.sum(np.min([np.zeros(xr_params.shape[0]-1), xr_params[1:,1] - xr_params[:-1,1]], axis=0)**2)
        discarded_ratio = max(MIN_LOG10_VALUE, (xr_scale - np.sum(ty[map_slice]**2))/xr_scale)
        unmapped_penalty = min(0, c)**2 + max(0, d - x[-1])**2 + max(-1.5, np.log10(discarded_ratio))

        if self.min_overlap:
            overlap = np.zeros(len(x))
            for i in range(len(cy_list)-1):
                cy1 = cy_list[i]
                cy2 = cy_list[i+1]
                overlap += np.abs(np.min([cy1, cy2], axis=0))       # np.abs() is intended to degrade negative elements
            overlap_penality = max(-1.5, np.log10(np.sum(overlap**2)/xr_scale))
        else:
            overlap_penality = 0

        scores = [max(resid_xr, resid_uv), resid_rg, mapping_penalty, vanishing_penalty, asymmetry_penalty, sigma_penalty, overlap_penality, disorder_penalty, unmapped_penalty]
        fv = np.sum(scores)

        if np.isnan(fv):
            # make clear the cause of this situation
            print("scores=", scores)
            print("xr_scale=", xr_scale)
            print("rg_scale=", rg_scale)
            print("uv_scale=", uv_scale)

        if debug_:
            from molass_legacy.Elution.CurveUtils import simple_plot

            disp_scores = [resid_xr, resid_uv, resid_rg, mapping_penalty, vanishing_penalty, asymmetry_penalty, sigma_penalty, overlap_penality, disorder_penalty, unmapped_penalty]

            dp = plt.push()
            fig, (ax1, ax2, ax3) = plt.subplots(ncols=3, figsize=(21,7))

            if fig_info is None:
                fig.suptitle("Simultaneous Optimization Result; fv=%.3g" % (fv), fontsize=24)
            else:
                in_folder, seeds, fv, show_scores = fig_info
                fig.suptitle("Simultaneous Optimization Result for %s with seeds=(%d, %d), fv=%.3g" % (in_folder, *seeds, fv), fontsize=24)
            ax1.set_title("UV Elution Decomposition", fontsize=20)
            ax2.set_title("Xray Elution Decomposition", fontsize=20)
            title = "Decomposition and Mapping Scores" if show_scores else "Mapped Elution Curves"
            ax3.set_title(title, fontsize=20)

            if False:
                def sort_components(vec_list):
                    sorted_list = sorted([(k, vec) for k, vec in enumerate(vec_list)], key=lambda x: xr_params[x[0],1])
                    return [rec[1] for rec in sorted_list]

                uv_cy_list = sort_components(uv_cy_list)
                cy_list = sort_components(cy_list)

            axt = ax2.twinx()
            axt.grid(False)

            simple_plot(ax1, self.uv_curve, color='blue', legend=False)
            simple_plot(ax2, self.xr_curve, color='orange', legend=False)

            k = 0
            for uv_cy, cy in zip(uv_cy_list, cy_list):
                k += 1
                ax1.plot(uv_x, uv_cy, ':', label='component-%d' % k)
                ax2.plot(x, cy, ':', label='component-%d' % k)

            ax1.plot(uv_x, uv_ty, ':', color='red', label='total')
            ax2.plot(x, ty, ':', color='red', label='total')

            axt.plot(xm, rg, ':', label='observed rg')
            axt.plot(xm, t_rg, color='gray', label='reconstructed rg')
            ymin, ymax = axt.get_ylim()
            axt.set_ylim(20, ymax)

            if self.min_overlap:
                y1 = np.zeros(len(x))
                ax2.fill_between(x, y1, overlap, fc='pink', alpha=0.2)

            if show_scores:
                y_pos = np.arange(len(disp_scores))
                ax3.barh(y_pos, disp_scores)
                ax3.set_yticks(y_pos)
                ax3.invert_yaxis() 
                ax3.set_yticklabels("resid_xr, resid_uv, resid_rg, mapping_penalty, vanishing_penalty, asymmetry_penalty, sigma_penalty, overlap_penality, disorder_penalty, unmapped_penalty".split(", "))
            else:
                ax3.plot(x, y, color='orange', label='Xray elution')
                uv_my = np.zeros(len(x))
                for cy in cy_list:
                    uv_my +=  cy/uv_ty * uv_y
                max_uv_my = np.max(uv_my)
                ax3.plot(x, uv_my/max_uv_my*self.xr_curve.max_y, ':', color='blue', label='mapped UV elution')

            def plot_mappable_range(ax, f, t):
                ymin, ymax = ax.get_ylim()
                ax.set_ylim(ymin, ymax)
                for x_ in [f, t]:
                    ax.plot([x_, x_], [ymin, ymax], color='yellow')

            u, v = [a*x_ + b for x_ in [c, d]]
            plot_mappable_range(ax1, u, v)
            plot_mappable_range(ax2, c, d)

            ax1.legend(fontsize=16)
            ax2.legend(loc='upper right', fontsize=16)
            axt.legend(loc='upper left', fontsize=16)
            ax3.legend(fontsize=16)

            fig.tight_layout()
            fig.subplots_adjust(top=0.85)
            if fig_info is None:
                self.debug_fv = plt.show()
            else:
                path = self.get_figure_path()
                dp.fig.canvas.draw()
                dp.fig.savefig(path)
                self.logger.info("xr_params=%s, rg_params=%s, (a, b)=%s, uv_params=%s, (c, d)=%s", 
                                    str(xr_params), str(rg_params), str((a,b)), str(uv_params), str((c, d)))
                # plt.show()
            plt.pop()

        return fv

    def get_figure_path(self):
        from molass_legacy._MOLASS.SerialSettings import get_setting
        sub_no = -1
        while True:
            sub_no += 1
            paren = '' if sub_no == 0 else ' (%d)' % sub_no
            path  = os.path.join(get_setting("temp_folder"), "fig-%03d.jpg%s" % (self.counter, paren))
            if os.path.exists(path):
                continue
            else:
                break
        return path

    def pretend(self, seeds, result, min_overlap=True):
        self.peaks_seed, self.bh_seed = seeds
        self.prepare_for_optimization(result.x)
        mean, v = self.xr_curve.compute_moments()
        s = np.sqrt(v)
        self.xr_sigma = s
        self.min_overlap = min_overlap

def spike_demo(in_folder, num_components=4, num_trials=1, seeds=None, result=None):
    from time import time
    import logging
    from molass_legacy.Batch.StandardProcedure import StandardProcedure
    from molass_legacy.Elution.CurveUtils import simple_plot
    from RgProcess.RgCurve import RgCurve
    from molass_legacy.SerialAnalyzer.DataUtils import get_in_folder

    logger = logging.getLogger(__name__)

    sp = StandardProcedure()
    sp.load(in_folder, debug=False)
    sd = sp.get_corrected_sd()

    D, E, qv, xr_curve = sd.get_xr_data_separate_ly()
    U, wlvec, uv_curve = sd.conc_array, sd.lvector, sd.get_uv_curve()

    t0 = time()
    rg_curve = RgCurve(qv, xr_curve, D, E)
    logger.info("It took %.3g seconds for rg_curve construction.", time()-t0)
    segments = rg_curve.get_curve_segments()

    solver = SimultaneousSolver(num_components, xr_curve, rg_curve, uv_curve)
    in_folder_ = get_in_folder(in_folder)

    init_seeds = np.random.randint(100000, 999999, size=num_trials)
    for k, init_seed in enumerate(init_seeds):
        logger.info("RandomState init_seed=%d", init_seed)
        np.random.RandomState(seed=init_seed)
        t0 = time()
        logger.info("solving with given seeds=%s.", str(seeds))
        if result is None:
            result = solver.solve(seeds=seeds, debug=EVAL_PARAMS_DEBUG)
        else:
            solver.pretend(seeds, result)
        used_seeds = solver.get_used_seeds()
        solver.objective_func(result.x, plot=True, fig_info=[in_folder_, used_seeds, result.fun, False])
        logger.info("It took %.3g seconds in Basin Hopping optimization with seeds=(%d, %d), fv=%.3g.", time()-t0, *used_seeds, result.fun)
        seeds = None
        result = None



"""
    UnifiedDecompResultTest.py

    Copyright (c) 2020, SAXS Team, KEK-PF
"""
import copy
import logging
from matplotlib.patches import Rectangle
import molass_legacy.KekLib.DebugPlot as plt

def unit_test(caller):
    from importlib import reload
    import UnifiedDecompResult
    reload(UnifiedDecompResult)
    from UnifiedDecompResult import UnifiedDecompResult

    editor = caller.dialog.get_current_frame()  # get editor because caller.editor is not updated by "Change Model"

    old_result = editor.decomp_result
    print("unit_test")
    # result = copy.deepcopy(old_result)
    result = UnifiedDecompResult(
                xray_to_uv=old_result.xray_to_uv,
                x_curve=old_result.x_curve,
                x=old_result.x,
                y=old_result.y,
                opt_recs=old_result.opt_recs,
                max_y_xray=old_result.max_y_xray,
                model_name=old_result.model_name,
                decomposer=old_result.decomposer,
                uv_y=old_result.uv_y,
                opt_recs_uv=old_result.opt_recs_uv,
                max_y_uv=old_result.max_y_uv,
                nresid_uv=old_result.nresid_uv,
                global_flag=old_result.global_flag,
                )

    logger = logging.getLogger(__name__)

    print("model_name=", result.model_name)
    result.remove_unwanted_elements()

    control_info = result.get_range_edit_info(logger=logger, debug=False)
    editor_ranges = control_info.editor_ranges
    print("editor_ranges=", editor_ranges)

    flags = result.identify_ignorable_elements()
    print("flags=", flags)

    with plt.Dp():
        x = result.x
        y = result.y
        fig, ax = plt.subplots()
        ax.set_title("UnifiedDecompResultTest")
        ax.plot(x, y)
        ymin, ymax = ax.get_ylim()
        ax.set_ylim(ymin, ymax )

        for list_ in editor_ranges:
            for f, t in list_:
                p = Rectangle(
                        (f, ymin),  # (x,y)
                        t - f,   # width
                        ymax - ymin,    # height
                        facecolor   = 'cyan',
                        alpha       = 0.2,
                    )
                ax.add_patch(p)

        fig.tight_layout()
        plt.show()



import sys
import os
import numpy                as np
import matplotlib.pyplot    as plt
# import statsmodels.api      as sm
from scipy                  import stats
sys.path.append( os.path.dirname( os.path.abspath( __file__ ) ) + '/../lib' )
import OurStatsModels       as sm


N   = 10
a   = 5
b   = 3
s   = 0.2
x   = np.linspace( 0, 1, N )
np.random.seed( 1234 )
e   = np.random.normal( 0, s, N )
yt  = a * x + b
y   = yt + e

X   = sm.add_constant( x )
model   = sm.WLS( y, X )
result  = model.fit()

print( 'result.params=', result.params )
print( 'result.cov_params=', np.sqrt( np.diag( result.cov_params() ) ) )

"""
# e2 = np.ones( (N,) ) * s**2
e2 =  ( y - np.dot( X, result.params ) )**2
print( 'e2=', e2 )
error = np.sqrt( np.dot( model.XtWX_inv_XtW**2, e2 ) )
print( 'error=', error )
"""

slope, intercept, _, _, std_err = stats.linregress( x, y )
print( 'intercept, slope=', [ intercept, slope ] )
print( 'std_err=', std_err )

param_array = []
for i in range(10000):
    e   = np.random.normal( 0, s, N )
    y   = yt + e
    model   = sm.WLS( y, X )
    result  = model.fit()
    param_array.append( result.params )

param_array_ = np.array( param_array )
print( 'std=', np.std( param_array_, axis=0 ) )



# coding: utf-8
"""
    PrerecogProxy.py

    Copyright (c) 2020-2021, SAXS Team, KEK-PF
"""
import logging
from molass_legacy.Trimming.PreliminaryRecognition import PreliminaryRecognition

class FlowChangeProxy:
    def __init__(self, uv):
        j_slice = uv.j_slice
        jmin = 0 if j_slice.start is None  else  j_slice.start
        jmax = uv.data.shape[1]-1 if j_slice.stop is None  else  j_slice.stop-1
        self.j_points = (jmin, jmax)

    def get_flow_changes(self):
        return self.j_points

class PrerecogProxy(PreliminaryRecognition):
    def __init__(self, md):
        self.logger = logging.getLogger(__name__)
        self.md = md
        self.cs = md.cs
        self.flowchange = FlowChangeProxy(md.uv)
        self.mapped_info = self.cs.mapped_info
        self.pre_rg = None

    def get_xray_shift(self):
        """
        overriding PreliminaryRecognition.get_xray_shift
        so that get_setting('xr_restrict_list') not be required.
        """
        xr = self.md.xr
        shift = xr.j_slice.start
        if shift is None:
            shift = 0
        return shift

    def get_rg(self):
        if self.pre_rg is None:
            self.sd = self.md.get_sd()
            self.get_default_angle_range(self.sd)
        return self.pre_rg.sg.Rg



"""
    _MOLASS/Processes.py

    Copyright (c) 2024, SAXS Team, KEK-PF
"""

processes = []

def register_process(process):
    processes.append(process)

def remove_process(process):
    process.join()
    processes.remove(process)

def terminate_all_processes():
    for process in processes:
        # task: do this more politely
        process.terminate()


"""
    SecTheory.ColumnTypeEstimator.py

    Copyright (c) 2023, SAXS Team, KEK-PF
"""
from .ColumnTypes import COLUMN_TYPE_DICT

class ColumnTypeEstimator:
    def __init__(self):
        pass

    def guess_from_sd(self, sd):
        pass

def guess_for_all_data():
    pass



"""
    Polynomial.py

    
"""
class poly1d:
    def __init__(self, coeffs):
        self.coeffs = coeffs

    def __call__(self, x):
        y = 0
        for c in self.coeffs:
            y = y*x + c
        return y



"""
    Optimizer.LrfExporter.py

    Copyright (c) 2022-2024, SAXS Team, KEK-PF
"""
import os
import re
import numpy as np
from molass_legacy.KekLib.NumpyUtils import np_savetxt
from molass_legacy._MOLASS.SerialSettings import get_setting
from molass_legacy.KekLib.BasicUtils import Struct
from molass_legacy.SerialAnalyzer.AnalyzerUtil import compute_conc_factor_util

class LrfExporter:
    def __init__(self, optimizer, params, dsets):
        xrD_ = optimizer.xrD_
        uvD_ = optimizer.uvD_
        self.optimizer = optimizer
        self.separate_eoii = optimizer.separate_eoii
        self.xr_curve = optimizer.xr_curve
        self.uv_curve = optimizer.uv_curve
        self.xrD = optimizer.xrD
        self.xrD_ = optimizer.xrD_
        self.xrE = optimizer.xrE
        self.uvD = optimizer.uvD
        self.uvD_ = optimizer.uvD_
        self.rg_curve = optimizer.rg_curve
        self.dsets = dsets
        self.params = params
        self.separate_params = optimizer.split_params_simple(params)
        self.conc_factor = compute_conc_factor_util()
        lrf_info = optimizer.objective_func(params, return_lrf_info=True)

        Pxr, Cxr, Puv, Cuv, mapped_UvD = lrf_info.matrices
        mapping = self.separate_params[3]
        uv_points = lrf_info.estimate_uv_peak_points()

        self.uv_result = Struct(Cuv=Cuv, Puv=Puv, uv_curve=self.uv_curve, mapping=mapping, peaktops=uv_points, uv_x=lrf_info.uv_x, uv_y=lrf_info.uv_y)

        xr_points = lrf_info.estimate_xr_peak_points()
        valid_rgs = lrf_info.get_valid_rgs(self.separate_params[2])
        self.xr_result = XrLrfResult(self.xrD, self.xrD_, self.xrE, Pxr, Cxr, optimizer.qvector,
                                    self.xr_curve, self.conc_factor, valid_rgs, xr_points)
                                    # self.xrD ok?

    def export(self, folder=None):
        from molass_legacy.KekLib.OurTkinter import Tk
        from molass_legacy.KekLib.PathUtils import NonExistingPath

        if folder is None:
            from .TheUtils import get_optimizer_folder
            optimizer_folder = get_optimizer_folder()
            folder = os.path.join(optimizer_folder, "exported")

        folder = str(NonExistingPath(folder))
        os.makedirs(folder)

        self.xr_result.export(folder)
        dummy_parent = Tk.Tk()  # this dummy_parent should be removed
        dummy_parent.withdraw()
        sheet = self.optimizer.params_type.get_params_sheet(dummy_parent, self.params, self.dsets, self.optimizer)
        sheet.save_as(os.path.join(folder, "parameters.csv"))
        dummy_parent.destroy()

        gr_result = self.get_graphic_result()
        gr_folder = os.path.join(folder, "graphic")
        os.makedirs(gr_folder)
        gr_result.export(gr_folder)

        return folder

    def get_graphic_result(self):
        return GraphicResult(self.uv_result, self.xr_result, self.rg_curve)

class XrLrfResult:
    def __init__(self, xrD, xrD_, xrE, P, C, qv, xr_curve, conc_factor, rg_params, peaktops):
        self.info = (xrD, xrD_, xrE, P, C, qv)
        self.xr_curve = xr_curve
        self.conc_factor = conc_factor
        self.rg_params = rg_params
        self.peaktops = peaktops

    def export(self, folder):
        
        """
        P  = M @ X, X = Minv @ P
        Pe <== E @ X
        Pe = sqrt((E**2) @ (X**2))
        """

        M, M_, E, P, C, qv = self.info

        X = np.linalg.pinv(M) @ P           # M or M_
        Ep = np.sqrt((E**2) @ (X**2))
        num_components = C.shape[0] - 1
        for j, p in enumerate(P.T):
            filename = 'component-%d.dat' % (j+1) if j < num_components else 'baseline-component.dat'
            file = os.path.join(folder, filename)
            np_savetxt(file, np.array([qv, p, Ep[:,j]]).T, fmt="%.5e")

        eno = np.arange(C.shape[1])
        for i, c in enumerate(C):
            filename = 'concentration-%d.dat' % (i+1) if i < num_components else 'baseline-concentration.dat'
            file = os.path.join(folder, filename)
            np_savetxt(file, np.array([eno, c]).T, fmt="%.5e")

NP_SAVETXT_FMT = "%g"

class GraphicResult:
    def __init__(self, uv_result, xr_result, rg_curve):
        self.uv_result = uv_result
        self.xr_result = xr_result
        self.rg_curve = rg_curve

    def export(self, folder):
        from GuinierTools.RgCurveUtils import compute_rg_curves

        xr_curve = self.xr_result.xr_curve
        xr_x = xr_curve.x
        xr_y = xr_curve.y
        C = self.xr_result.info[4]  # C
        CY = C/self.xr_result.conc_factor

        # xr
        file = os.path.join(folder, "xr-elution-data.txt")
        np.savetxt(file, np.array([xr_x, xr_y]).T, fmt=NP_SAVETXT_FMT)
        ty = np.zeros(len(xr_x))
        xr_cy_list = []
        for k, cy in enumerate(CY, start=1):
            if k < len(CY):
                name = "xr-elution-c-%d.txt" % k
            else:
                name = "xr-elution-baseline.txt"
            file = os.path.join(folder, name)
            np.savetxt(file, np.array([xr_x, cy]).T, fmt=NP_SAVETXT_FMT)
            ty += cy
            xr_cy_list.append(cy)

        file = os.path.join(folder, "xr-elution-c-total.txt")
        np.savetxt(file, np.array([xr_x, ty]).T, fmt=NP_SAVETXT_FMT)
        peaktops = self.xr_result.peaktops
        file = os.path.join(folder, "xr-peaktop-dots.txt")
        np.savetxt(file, peaktops, fmt=NP_SAVETXT_FMT)

        # rg
        ty_ = ty - xr_cy_list[-1]       # remove baseline part
        weights = self.xr_result.peaktops[:,1]
        rg_params = self.xr_result.rg_params
        rg_curves1, rg_curves2 = compute_rg_curves(xr_x, weights/np.max(weights), rg_params, xr_cy_list[:-1], ty_, self.rg_curve)
        for k, ((x1, rg1), (x2, rg2)) in enumerate(zip(rg_curves1, rg_curves2), start=1):
            file = os.path.join(folder, "rg-observed-%d.txt" % k)
            np.savetxt(file, np.array([x1, rg1]).T, fmt=NP_SAVETXT_FMT)
            file = os.path.join(folder, "rg-reconstructed-%d.txt" % k)
            np.savetxt(file, np.array([x2, rg2]).T, fmt=NP_SAVETXT_FMT)

        # uv
        uv_x = self.uv_result.uv_x
        uv_y = self.uv_result.uv_y
        file = os.path.join(folder, "uv-elution-data.txt")
        np.savetxt(file, np.array([uv_x, uv_y]).T, fmt=NP_SAVETXT_FMT)

        Cuv = self.uv_result.Cuv
        ty = np.zeros(len(uv_x))
        for k, cy in enumerate(Cuv, start=1):
            if k < len(Cuv):
                name = "uv-elution-c-%d.txt" % k
            else:
                name = "uv-elution-baseline.txt"
            file = os.path.join(folder, name)
            np.savetxt(file, np.array([uv_x, cy]).T, fmt=NP_SAVETXT_FMT)
            ty += cy

        file = os.path.join(folder, "uv-elution-c-total.txt")
        np.savetxt(file, np.array([uv_x, ty]).T, fmt=NP_SAVETXT_FMT)
        peaktops = self.uv_result.peaktops
        file = os.path.join(folder, "uv-peaktop-dots.txt")
        np.savetxt(file, peaktops, fmt=NP_SAVETXT_FMT)



# coding: utf-8
"""
    KekToolsGP.py

    Copyright (c) 2016-2017, Masatsuyo Takahashi, KEK-PF
"""

import sys
import logging
import traceback
import time
import numpy                as np
from bisect                 import bisect_right
from molass_legacy.KekLib.ExceptionTracebacker   import ExceptionTracebacker
from IntensityData          import IntensityData, ACCEPTABLE_BASIC_QUALITY
from GuinierPorodFit        import GuinierPorodFit
from Guinier                import Guinier, AIC_SCORE_BOUNDARY
from GuinierInterval        import GuinierInterval
from QrgLimitsAdjuster      import QrgLimitsAdjuster
from Quality                import is_better_quality, fit_score_func
from ErrorResult            import ErrorResult

DEBUG = False

"""
    AutorgKek   ->  GuinierPorodFit     fitting
                    SmoothLineGP        fit result in x, y
                    GuinierInterval     optimization
"""

class AutorgKek:
    def __init__( self, file=None ):
        self.logger = logging.getLogger( __name__ )

        if file is None: return

        if type( file ) == str or type( file ) == np.ndarray:
            try:
                intensity = IntensityData( file, add_smoother=True )
            except ( ValueError, IndexError ):
                # these errors are known for very low quality data
                # which include many negative values.
                etb = ExceptionTracebacker()
                self.logger.error( etb )
                intensity = None
        elif type( file ) == IntensityData:
            intensity = file
        else:
            assert( False )

        self.intensity = intensity
        self.interval = None
        self.optimal_interval_q = None

    def run( self, robust=False, optimize=True,
                qrg_limits=[ 0.0, 1.3 ],
                retry_mode=False
           ):

        if DEBUG: print( 'AutorgKek.run: qrg_limits=', qrg_limits )

        if self.intensity is None or not self.intensity.has_acceptable_quality():
            self.result = ErrorResult()
            return self.result

        better_fit = None
        force_low_quality = False

        self.fit = fit = GuinierPorodFit( self.intensity )
        fit.fit()

        x, y, e = self.intensity.get_guinier_valid_xy()
        self.guinier = Guinier( x, y, e, positive_ratio=self.intensity.positive_ratio )

        # boundary_q = qrg_limits[1] / fit.Rg * np.sqrt( 3*fit.degree/2 )
        boundary_q = qrg_limits[1] / fit.Rg

        if DEBUG: print( 'basic_quality=', self.intensity.basic_quality )

        is_sufficient_baisic_quality = self.intensity.basic_quality >= ACCEPTABLE_BASIC_QUALITY and not force_low_quality

        if is_sufficient_baisic_quality:
            allow_ratio_list    = [ 1 ]
        else:
            allow_ratio_list    = [ 1, 1.2, 1.5, 4.0 ]

        for allow_ratio in allow_ratio_list:
            if DEBUG: print( '[0] --------------------------------- without interval optimization: allow_ratio=', allow_ratio )

            t_ = bisect_right( self.intensity.Q, boundary_q * allow_ratio )

            # print( 'boundary_q=', boundary_q, '; fit.f0, t_=', [ fit.f0, t_  ] )

            """
            if is_sufficient_baisic_quality:
                f0 = fit.f0
            else:
                f0 = 0
            """

            f0 = fit.f0

            if t_ > f0:
                try:
                    self.guinier.estimate_rg( [ f0, t_ ], qrg_limits, fit.Rg )
                    result1 = self.guinier.get_result( fit, result_type=1 )
                    if DEBUG: print( '[ f0, t_ ]=', [ f0, t_ ], ', result1: Rg=', result1.Rg, ', Quality=', result1.Quality )
                except Exception as e:
                    print( e )
                    result1 = None
            else:
                result1 = None

            if result1 is not None and result1.Rg is not None:
                break

        if DEBUG: print( 'result1=', result1 )

        if is_sufficient_baisic_quality:
            opt_result  = None
            opt_quality = None
            boundary_q = qrg_limits[1] / fit.Rg
            # allow_ratio_list    = [ 0.8, 0.9, 1 ]
            allow_ratio_list    = [ 0.9, 1, 1.1, 1.2, 1.5, 2.0 ]

            for i, allow_ratio in enumerate( allow_ratio_list ):
                if DEBUG: print( '[1] --------------------------------- with interval optimization: allow_ratio=', allow_ratio )
                f0 = fit.f0
                t_ = bisect_right( self.intensity.Q, boundary_q * allow_ratio )
                # print( 'f0, t_=', [ f0, t_ ] )
                if f0 >= t_:
                    continue

                try:
                    slice_ = slice( f0, t_+1 )
                    gx = self.intensity.Q[slice_]
                    gy = self.intensity.I[slice_]
                    gw = self.intensity.W[slice_]
                    self.interval = GuinierInterval( self.intensity.Q, self.intensity.I, gx, gy, gw, fit, qrg_limits )
                    min_q, max_q, f_, t_ = self.interval.get_optimal_interval()
                    if DEBUG: print( 'f0, t_=', ( f0, t_ ), '; f_, t_=', ( f_, t_ ) )

                    self.guinier.estimate_rg( [ f0, t_ ], qrg_limits, fit.Rg )
                    result_ = self.guinier.get_result( fit, result_type=2 )
                    if result_ is None or result_.Rg is None:
                        continue

                    self.optimal_interval_q = [ min_q, max_q ]
                    if DEBUG: print( 'optimization is ok. Rg=%.4g, Quality=%.4g, max_qRg=%.4g' % ( result_.Rg, result_.Quality, result_.max_qRg ) )

                except Exception as e:
                    if True:
                        ( exc_type_type_, exc_value, exc_traceback ) = sys.exc_info()
                        e_seq = traceback.format_exception( exc_type_type_, exc_value, exc_traceback )
                        print( 'Optimization Error! ' + ''.join( e_seq ) )
                    result_ = None
                    continue

                if self.intensity.basically_ok:
                    # evaluate quality without q_rg_score
                    if result_ is None:
                        quality_ = 0
                    else:
                        quality_ = np.sum( result_.quality_object.get_factors()[0:5] )

                    if opt_quality is None or quality_ > opt_quality:
                        opt_quality = quality_
                        opt_result  = result_
                else:
                    if opt_result is None or is_better_quality( result_, opt_result, qrg_limits, allow_qrg=0.1 ):
                        # allow_qrg=0.1 makes 'AIMdim01_00158_sub.dat' slightly better ( 21.7 => 23.9 )
                        opt_result = result_
                        # print( 'opt_result.Rg=', opt_result.Rg )

                if i == 4 and opt_result is not None and opt_result.Rg is not None:
                    break

            result2 = opt_result
        else:
            result2 = None

        if result2 is None:
            result2 = self.guinier.get_result( fit, result_type=4 )

        if True:
            if result1 is None:
                self.result = result2
            else:
                if self.intensity.basically_ok:
                    if DEBUG: print( 'fit_consistency1=',  result1.quality_object.fit_consistency_pure)
                    if DEBUG: print( 'fit_consistency2=',  result2.quality_object.fit_consistency_pure)
                    if result2.quality_object.fit_consistency_pure > 0.3:
                        self.result = result2
                    else:
                        if result1.quality_object.fit_consistency_pure > result2.quality_object.fit_consistency_pure:
                            self.result = result1
                        else:
                            self.result = result2
                else:
                    if is_better_quality( result2, result1, qrg_limits ):
                        self.result = result2
                    else:
                        self.result = result1
        else:
            if is_better_quality( result2, result1, qrg_limits ):
                self.result = result2
            else:
                self.result = result1

        assert( qrg_limits is not None )
        try:
            # if self.result is not None: print( 'before adjust result.Rg=', self.result.Rg )

            self.result = self.adjust_to_qrg_limits( qrg_limits, self.result )

            # if self.result is not None: print( 'adjusted result.Rg=', self.result.Rg )
        except Exception as e:
            # when result.Rg is None
            if DEBUG:
                ( exc_type_type_, exc_value, exc_traceback ) = sys.exc_info()
                e_seq = traceback.format_exception( exc_type_type_, exc_value, exc_traceback )
                print( ''.join( e_seq ) )
            pass

        if self.result is None:
            self.result = ErrorResult()
        else:
            self.result = self.guinier.update_result_with_bico_solver( self.result )

        if not retry_mode and self.result.IpI == 1:
            # print( 'retry for IpI' )
            # TODO: 
            start_ipi_retry = self.guinier.determine_ipi_interval()
            # print( 'start_ipi_retry=', start_ipi_retry )

            intensity_for_retry = self.intensity.copy( start=start_ipi_retry )
            retry_obj = AutorgKek( intensity_for_retry )
            retried_result = retry_obj.run( retry_mode=True )
            # print( 'retried_result.Rg=', retried_result.Rg, ', retried_result.Aggregated=', retried_result.Aggregated )

            if retried_result.IpI != 0:
                # keep the low quality as IpI
                retried_result.quality_object.degrade_by_anomalies( 0.5 )
            head_cvmin = self.result.head_cvmin
            self.result = retried_result
            self.result.IpI         = 1
            self.result.head_cvmin  = head_cvmin
            self.result.From        += start_ipi_retry
            self.result.To          += start_ipi_retry

        self.result.Rg_ = 0 if self.result.Rg is None else self.result.Rg
        self.result.I0_ = 0 if self.result.I0 is None else self.result.I0

        return self.result

    def adjust_to_qrg_limits( self, qrg_limits, result, ):
        adjuster = QrgLimitsAdjuster( self.fit, self.guinier, qrg_limits, result )
        return adjuster.get_adjusted_result()

    def proof_plot( self, plot_class, title=None, sleep=0 ):
        plot = plot_class( self.fit, self.interval, self.guinier,  self.optimal_interval_q, title=title )
        block_ = sleep == 0
        plot.show( block=block_ )
        if sleep > 0:
            time.sleep( sleep )
        plot.close()

def autorg( file, robust=False, optimize=True,
            qrg_limits=[ 0, 1.3 ], qrg_limits_apply=True ):
    autorg_ = AutorgKek( file  )
    autorg_.run( robust=robust, optimize=optimize,
                    qrg_limits=qrg_limits )
    return autorg_.result, autorg_.intensity



"""
    Optimizer.FullOptDialog.py

    Copyright (c) 2021-2024, SAXS Team, KEK-PF
"""
import sys
import os
import numpy as np
import scipy.stats as stats
import re
import logging
from importlib import reload
from molass_legacy.KekLib.BasicUtils import ordinal_str
from molass_legacy.KekLib.OurTkinter import Tk, Dialog, ttk
from molass_legacy.KekLib.TkCustomWidgets import FolderEntry
from molass_legacy._MOLASS.Version import get_version_string, is_developing_version
import molass_legacy.KekLib.CustomMessageBox as MessageBox
from molass_legacy.KekLib.ExceptionTracebacker import log_exception
from molass_legacy._MOLASS.SerialSettings import get_setting
from .OptConstants import MIN_NUM_COMPONENTS, MAX_NUM_COMPONENTS
from .BackRunner import BackRunner
from .JobStateInfo import JobStateInfo
from .OptJobInfo import OptJobInfo, JOB_STATES
from .OptJobResultInfo import OptJobResultInfo
from molass_legacy.Peaks.PeakEditor import DRIFT_TYPES
from .OptimizerSettings import get_advanced_settings_text
from .ElutionComposer import COMPOSER_CB_TEXT

CANVAS_DEBUG = False
POLL_INTERVAL = 1000
# DEVELOP_MODE = sys.executable.find("pythonw") < 0
DEVELOP_MODE = is_developing_version()

class FullOptDialog(Dialog):
    def __init__(self, parent, parent_dialog, optinit_info, **kwargs):
        self.logger = logging.getLogger(__name__)
        self.parent = parent
        self.parent_dialog = parent_dialog

        self.fullopt = None
        self.runner = BackRunner()
        self.optinit_info = optinit_info
        self.class_code = optinit_info.class_code
        self.param_init_type = optinit_info.param_init_type
        self.bounds_type = optinit_info.bounds_type     # not used yet
        # self.job_list = kwargs.pop('job_list', [])
        self.job_list = optinit_info.job_list
        self.demo_index = kwargs.pop('demo_index', 0)
        self.inspecting = kwargs.pop('inspecting', False)   # True for Result View
        self.kwargs = kwargs
        self.terminating = False
        self.attending_current = True
        self.is_busy_ = not self.inspecting
        self.result_list = []
        self.known_best_fv = None
        self.known_best_index = None
        self.state_info = JobStateInfo(optinit_info, self.demo_index)
        self.fullopt = self.state_info.fullopt
        self.init_info = self.state_info.get_init_info()
        self.elution_model = get_setting("elution_model")
        self.recompute_rg_curve = get_setting("recompute_rg_curve")
        pid = os.getpid()
        version = get_version_string()
        self.poll_using_filesize = False     # temporary fix for the shared memory bug
        Dialog.__init__(self, parent, "Optimization Monitor - %s pid=%d" % (version, pid), visible=False)

    def initialize(self):
        pass

    def get_skipbutton_name(self):
        return "Skip" if self.is_busy_ else "Resume"

    def is_busy(self):
        return self.is_busy_

    def show(self):
        if self.inspecting:
            init_method = self.set_inspection_mode
        else:
            init_method = self.run
        self.is_focused = True
        self.after(100, init_method)
        self._show()

    def body(self, body_frame):
        if CANVAS_DEBUG:
            import molass_legacy.Optimizer.JobStateCanvas
            reload(molass_legacy.Optimizer.JobStateCanvas)
        from .JobStateCanvas import JobStateCanvas
        self.canvas = JobStateCanvas(body_frame, self)
        self.canvas.pack()
        self.num_params = 22 if self.init_info is None else self.init_info[1].shape[1]
        self.logger.info("num_params=%d", self.num_params)

        self.num_jobs = Tk.IntVar()
        self.folder_box = None

        if len(self.job_list) == 0:
            self.folders = []           # used in add_new_job
            self.add_new_job(init=True)
            self.curr_job = 0
        else:
            self.folders = [job_info.name for job_info in self.job_list]
            job_name = self.kwargs.pop('job_name')
            self.curr_job = int(job_name)

        num_jobs = len(self.job_list)
        self.num_jobs.set(num_jobs)

        self.canvas.draw_main()

        pframe = Tk.Frame(body_frame)
        pframe.pack(padx=20)
        self.build_panel(pframe)

    def get_job_info(self):
        return self.job_list[self.curr_job]

    def build_panel(self, pframe):
        curr_job_info = self.job_list[self.curr_job]
        print("curr_job_info=", curr_job_info)

        space_height = 5

        grid_row = 0
        space =Tk.Frame(pframe, height=space_height)
        space.grid(row=grid_row, column=0)

        grid_row += 1
        label = Tk.Label(pframe, text="Job Root Folder: ")
        label.grid(row=grid_row, column=0, sticky=Tk.E)
        self.optjob_folder = Tk.StringVar()
        self.optjob_folder.set(self.runner.get_optjob_folder())

        self.root_folder_entry = FolderEntry(pframe, textvariable=self.optjob_folder, width=100,
            on_entry_cb=self.on_entry_root_folder)
        self.root_folder_entry.grid(row=grid_row, column=1, columnspan=11, sticky=Tk.W)
        self.root_folder_entry.config(state=Tk.DISABLED)    # 
 
        grid_row += 1
        space =Tk.Frame(pframe, height=space_height)
        space.grid(row=grid_row, column=0)

        grid_row += 1
        label = Tk.Label(pframe, text="Max Number of Trials: ")
        label.grid(row=grid_row, column=0, sticky=Tk.E)

        self.maxnum_trials = Tk.IntVar()
        self.maxnum_trials.set(self.optinit_info.maxnum_trials)
        num_jobs = self.num_jobs.get()
        self.mn_sbox = Tk.Spinbox(pframe, textvariable=self.maxnum_trials,
                          from_=num_jobs, to=999, increment=1,
                          justify=Tk.CENTER, width=6)
        self.mn_sbox.grid(row=grid_row, column=1, sticky=Tk.W)
        self.maxnum_trials.trace("w", self.maxnum_trials_tracer)

        space_width = 20
        grid_row += 1
        space =Tk.Frame(pframe, height=space_height)
        space.grid(row=grid_row, column=0)

        grid_row += 1
        grid_col = 0
        label = Tk.Label(pframe, text="Number of Jobs: ")
        label.grid(row=grid_row, column=grid_col, sticky=Tk.E)

        grid_col += 1
        label = Tk.Label(pframe, textvariable=self.num_jobs, bg='white', width=3)
        label.grid(row=grid_row, column=grid_col, sticky=Tk.W)

        grid_col += 1
        space = Tk.Frame(pframe, width=space_width)
        space.grid(row=grid_row, column=grid_col)

        grid_col += 1
        label = Tk.Label(pframe, text="Attending Job: ")
        label.grid(row=grid_row, column=grid_col)

        grid_col += 1
        self.curr_folder = Tk.StringVar()
        self.curr_folder.set(curr_job_info[0])
        self.folder_box = ttk.Combobox(master=pframe, values=self.folders, textvariable=self.curr_folder,
                                width=5, justify=Tk.CENTER, state=Tk.DISABLED)
        self.folder_box.grid(row=grid_row, column=grid_col)
        self.curr_folder.trace("w", self.curr_folder_tracer)

        grid_col += 1
        space = Tk.Frame(pframe, width=space_width)
        space.grid(row=grid_row, column=grid_col)

        grid_col += 1
        label = Tk.Label(pframe, text="Job State: ")
        label.grid(row=grid_row, column=grid_col)

        grid_col += 1
        self.job_state = Tk.StringVar()
        self.job_state.set(curr_job_info[1])
        label = Tk.Label(pframe, textvariable=self.job_state, bg='white')
        label.grid(row=grid_row, column=grid_col)

        grid_col += 1
        space = Tk.Frame(pframe, width=5)
        space.grid(row=grid_row, column=grid_col)

        grid_col += 1
        label = Tk.Label(pframe, text="PID: ")
        label.grid(row=grid_row, column=grid_col)

        grid_col += 1
        self.job_pid = Tk.StringVar()
        label = Tk.Label(pframe, textvariable=self.job_pid, bg='white')
        label.grid(row=grid_row, column=grid_col)

        grid_col += 1
        space = Tk.Frame(pframe, width=space_width)
        space.grid(row=grid_row, column=grid_col)

        grid_col += 1
        label = Tk.Label(pframe, text="Number of Components: ")
        label.grid(row=grid_row, column=grid_col)

        grid_col += 1
        self.n_components = Tk.IntVar()
        # self.n_components.set(curr_job_info[2])
        self.n_components.set(self.fullopt.n_components - 1)    # just for display to the user
        self.nc_sbox = Tk.Spinbox(pframe, textvariable=self.n_components,
                          from_=MIN_NUM_COMPONENTS, to=MAX_NUM_COMPONENTS, increment=1,
                          justify=Tk.CENTER, width=6, state=Tk.DISABLED)
        self.nc_sbox.grid(row=grid_row, column=grid_col)

        grid_col += 1
        space = Tk.Frame(pframe, width=space_width)
        space.grid(row=grid_row, column=grid_col)

        grid_col += 1
        label = Tk.Label(pframe, text="Baseline: ")
        label.grid(row=grid_row, column=grid_col)

        label_text = get_advanced_settings_text()
        if label_text > "":
            label = Tk.Label(pframe, text=label_text)
            label.grid(row=grid_row-2, column=grid_col-2, columnspan=6)

        grid_col += 1
        unified_baseline_type = get_setting("unified_baseline_type")
        self.drift_type = Tk.StringVar()
        self.drift_type.set(DRIFT_TYPES[unified_baseline_type - 1])
        self.drift_type_box = ttk.Combobox(master=pframe,
                    values=DRIFT_TYPES, textvariable=self.drift_type, width=15, state=Tk.DISABLED)
        self.drift_type_box.grid(row=grid_row, column=grid_col)

        grid_col += 1
        space = Tk.Frame(pframe, width=space_width)
        space.grid(row=grid_row, column=grid_col)

        grid_col += 1
        label = Tk.Label(pframe, text="Number of Iterations: ")
        label.grid(row=grid_row, column=grid_col)

        grid_col += 1
        self.num_iter = Tk.IntVar()
        self.num_iter.set(curr_job_info[4])
        self.num_iter_entry = Tk.Entry(pframe, textvariable=self.num_iter, width=5, justify=Tk.CENTER)
        self.num_iter_entry.grid(row=grid_row, column=grid_col)

        grid_col += 1
        space = Tk.Frame(pframe, width=space_width)
        space.grid(row=grid_row, column=grid_col)

        grid_col += 1
        label = Tk.Label(pframe, text="Seed: ")
        label.grid(row=grid_row, column=grid_col)

        grid_col += 1
        self.seed = Tk.IntVar()
        self.seed.set(curr_job_info[5])
        self.seed_entry = Tk.Entry(pframe, textvariable=self.seed, width=10, justify=Tk.CENTER)
        self.seed_entry.grid(row=grid_row, column=grid_col)

        if is_developing_version():
            button = Tk.Button(pframe, text="Devel Test", command=self.devel_test)
            # note that 4 rows here above include 2 space rows.
            button.grid(row=grid_row-4, rowspan=4, column=grid_col, sticky=Tk.E)
 
        grid_row += 1
        space =Tk.Frame(pframe, height=space_height)
        space.grid(row=grid_row, column=0)

        grid_row += 1
        grid_col = 1
        self.try_model_composing = Tk.IntVar()
        self.try_model_composing.set(get_setting('try_model_composing'))
        cb = Tk.Checkbutton(pframe, text=COMPOSER_CB_TEXT, variable=self.try_model_composing)
        cb.grid(row=grid_row, column=grid_col, columnspan=5, sticky=Tk.W)
        self.try_model_composing_cb = cb

        grid_row += 1
        space =Tk.Frame(pframe, height=space_height)
        space.grid(row=grid_row, column=0)

    def buttonbox(self):
        box = Tk.Frame(self)
        box.pack()

        show_modify_button = DEVELOP_MODE and self.elution_model == 0
        padx = 90 if show_modify_button else 120

        w = Tk.Button(box, text="◀ Close", width=10, command=self.cancel)
        w.pack(side=Tk.LEFT, padx=padx, pady=10)
        self.cancel_button = w

        if show_modify_button:
            in_folder = get_setting("in_folder")
            state = Tk.NORMAL if in_folder.find("HasA") >= 0 else Tk.DISABLED
            w = Tk.Button(box, text="▽ Modify Restart", width=20, command=self.modify_restart, state=state)
            w.pack(side=Tk.LEFT, padx=padx, pady=10)
            self.modify_button = w

        name = self.get_skipbutton_name()
        w = Tk.Button(box, text=name, width=10, command=self.skip_to_next_trial)
        w.pack(side=Tk.LEFT, padx=padx, pady=10)
        self.skip_button = w
        self.remaining_dependents = [self.skip_button]

        w = Tk.Button(box, text="Terminate", width=12, command=self.terminate,
                            state=Tk.DISABLED)
        w.pack(side=Tk.LEFT, padx=padx, pady=10)
        self.stop_button = w
        self.stop_button_configs = [w.cget(a) for a in ['fg', 'bg']]

        w = Tk.Button(box, text="▶ Export", width=12, command=self.export)
        w.pack(side=Tk.LEFT, padx=padx, pady=10)

        self.job_state_dependents = [self.nc_sbox, self.seed_entry, self.num_iter_entry]
        self.bind("<Escape>", self.cancel)

        # do this later because it takes a while for those states getting ready
        self.after(500, self.update_try_model_composing_cb)

    def update_skipbutton_name(self):
        button_name = self.get_skipbutton_name()
        self.skip_button.config(text=button_name)

    def cancel(self):
        print("FullOptDialog.cancel")
        if self.is_busy_:
            MessageBox.showinfo("Can't Close Notification",
                "There seems to remain a still running job.\n"
                'You have to wait until the finish or "Terminate" before close.',
                parent=self)
            return

        self.canvas.close()
        Dialog.cancel(self)

    def run(self, restart_params=None, real_bounds=None, debug=True):
        from .FullOptResult import FILES, OPTIONAL_FILES

        assert self.job_state.get() == JOB_STATES[0]
        self.is_busy_ = True
        self.update_skipbutton_name()
        self.update_count = 0

        for w in self.job_state_dependents:
            w.config(state=Tk.DISABLED)

        if restart_params is None:
            fv_array, x_array = self.state_info.get_init_info()[0:2]
            init_params = x_array[0]
            if self.param_init_type == 0:
                # foreground plot has been initialized in the JobStateCanvas' initialization
                self.logger.info("got init_params from the initialization: num_params=", len(init_params))
            else:
                init_params = self.get_known_best_params(init_params)
        else:
            self.logger.info("resuming with restart_params=%s", restart_params)
            init_params = restart_params

        # try model composing
        # note that this mus be done before prepare_init_state
        work_folder = self.runner.get_work_folder()
        composite_info_txt = os.path.join(work_folder, OPTIONAL_FILES[0])   # composite_info.txt

        optimizer = self.fullopt
        try_model_composing = self.try_model_composing.get()
        composite = optimizer.composite
        if composite.really_composite:
            composite.save(composite_info_txt)
        else:
            if try_model_composing and len(self.job_list) > 1:
            # if try_model_composing:
                if debug:
                    from importlib import reload
                    import Optimizer.ElutionComposer
                    reload(Optimizer.ElutionComposer)
                from .ElutionComposer import ElutionComposer
                composer = ElutionComposer(optimizer, init_params, self.optinit_info.sd)
                if composer.reducible():
                    new_composite = composer.make_composite()
                    new_composite.save(composite_info_txt)
                    optimizer.set_composite(new_composite)

        # prepare for initial state
        if self.param_init_type == 0:
            pass
        else:
            # foreground plot must be synchronized with the background job
            self.state_info.prepare_init_state(init_params=init_params, real_bounds=real_bounds)
            self.canvas.update_init_state(self.state_info)

        niter = self.num_iter.get()
        seed = self.seed.get()

        # note that self.try_model_composing may have been changed when resuming from the "Result View"
        try_model_composing = self.try_model_composing.get()

        self.runner.run(optimizer, init_params, niter=niter, seed=seed, work_folder=work_folder)
        self.job_state.set(JOB_STATES[1])
        self.job_list[self.curr_job].update_state(JOB_STATES[1])
        self.stop_button.config(fg='white', bg='red', state=Tk.NORMAL)

        self.np_shm = self.runner.np_shm
        if self.np_shm is None:
            self.poll_using_filesize = True
        if self.poll_using_filesize:
            self.last_filesize = None
        else:
            self.last_callback_counter = None

        self.callback_txt_path = self.runner.get_callback_txt_path()
        self.after(POLL_INTERVAL, self.poll)

    def add_new_job(self, init=False):
        num_jobs = len(self.job_list)
        job_name = "%03d" % num_jobs
        job_state = JOB_STATES[0]

        if init:
            drift_type = DRIFT_TYPES[0]
            try:
                niter = self.optinit_info.n_iterations
                seed = self.optinit_info.seed
            except:
                niter = 100
                seed = np.random.randint(100000, 999999)
            n_components = int((self.num_params - 4)/6)
        else:
            n_components = self.n_components.get()
            drift_type = self.drift_type.get()
            niter = self.num_iter.get()
            seed = np.random.randint(100000, 999999)

        self.job_list.append(OptJobInfo(job_name, job_state, n_components, drift_type, niter, seed))
        self.folders.append(self.job_list[-1][0])

        num_jobs = len(self.job_list)
        self.num_jobs.set(num_jobs)

        if not init:
            self.folder_box.config(values=self.folders)
            self.mn_sbox.config(from_=num_jobs)

        if not init:
            num_remaining = self.maxnum_trials.get() - self.num_jobs.get()
            if num_remaining == 0:
                for w in self.remaining_dependents:
                    w.config(state=Tk.DISABLED)

    def poll(self):
        if self.is_busy_:
            ret = self.runner.poll()
            if ret is None:
                self.after(POLL_INTERVAL, self.poll)
                if self.job_pid.get() == "":
                    self.job_pid.set(str(self.runner.getpid()))
            else:
                self.logger.info("job %s finished with returncode=%d after %d state updates.", self.job_list[-1][0], ret, self.update_count)
                self.inspecting = True      # in order to enable "Resume", 
                # consider whether self.set_inspection_mode() is required
                self.job_pid.set("")
                job_state = JOB_STATES[2] if ret == 0 else JOB_STATES[3]
                self.job_state.set(job_state)
                self.job_list[self.curr_job].update_state(job_state)
                self.save_the_result_figure()
                self.try_update_information(append=True)    # append the result info before adding a new job
                num_remaining = self.maxnum_trials.get() - self.num_jobs.get()
                if num_remaining > 0 and not self.terminating:
                    self.start_next_trial()
                else:
                    self.curr_folder_tracer()
                    self.is_busy_ = False
                    self.update_skipbutton_name()
                    if num_remaining == 0:
                        for w in self.remaining_dependents:
                            w.config(state=Tk.DISABLED)
            self.try_update_information()

    def save_the_result_figure(self, fig_file=None):
        if fig_file is None:
            from .TheUtils import get_optimizer_folder
            optimizer_folder = get_optimizer_folder()
            figs_folder = os.path.join(optimizer_folder, "figs")
            if not os.path.exists(figs_folder):
                os.makedirs(figs_folder)
            fig_file = os.path.join(figs_folder, "fig-%s.jpg" % self.job_list[-1][0])

        self.canvas.save_the_figure(fig_file)

    def start_next_trial(self, restart_params=None, real_bounds=None):
        num_jobs = len(self.job_list)
        job_name = "%03d" % num_jobs
        if self.recompute_rg_curve:
            import RgProcess.RgCurveComputer
            reload(RgProcess.RgCurveComputer)
            from RgProcess.RgCurveComputer import RgCurveComputerDialog
            dialog = RgCurveComputerDialog(self, self, job_name)
            dialog.show()
            if not dialog.applied:
                return

        self.add_new_job()
        self.curr_folder.set(job_name)
        self.update()   # this will end up to call curr_folder_tracer
        self.run(restart_params=restart_params, real_bounds=real_bounds)

    def try_update_information(self, append=False, debug=False):
        if self.poll_using_filesize:
            try:
                filesize = os.path.getsize(self.callback_txt_path)
            except:
                # probably not yet exist
                timestamp = None
                filesize = 0

            updated = ( self.last_filesize is None
                        or filesize != self.last_filesize
                        )
            if debug:
                self.logger.info("try_update_information: filesize=%s, updated=%s;  ",      # ";  " has been added to avoid log overwrite confusion
                              filesize, updated)
        else:
            callback_counter = self.np_shm.array[0]
            updated = ( self.last_callback_counter is None
                        or callback_counter != self.last_callback_counter
                        )
            if debug:
                self.logger.info("try_update_information: callback_counter=%d, updated=%s;  ",      # ";  "  has been added to avoid log overwrite confusion
                             callback_counter, updated) 

        if (append      # make sure to pass the best result to the next trial
            or updated):
            try:
                self.update_information(append=append)
            except:
                log_exception(self.logger, "update_information: ")

        if self.poll_using_filesize:
            self.last_filesize = filesize
        else:
            self.last_callback_counter = callback_counter

    def update_information(self, work_folder=None, append=False, debug=True):
        if work_folder is None:
            work_folder = self.get_curr_work_folder()
        cb_file = os.path.join(work_folder, 'callback.txt')
        fv_list, x_list = self.read_callback_txt(cb_file)
        if debug:
            self.logger.info("updating information from %s, len(x_list)=%d", cb_file, len(x_list))
        xmax = self.estimate_xmax(fv_list)
        seed = self.seed.get()
        demo_info = np.array(fv_list), np.array(x_list), xmax, seed
        self.canvas.set_demo_info(demo_info)
        if append:
            self.append_result(demo_info)
        self.update_count += 1

    def get_curr_work_folder(self):
        optjob_folder = self.runner.get_optjob_folder()
        curr_folder = self.curr_folder.get()
        return os.path.join(optjob_folder, curr_folder)

    def estimate_xmax(self, fv_list):
        niter = self.num_iter.get()
        counter = fv_list[-1][0]
        if counter == 0:
            # init state
            xmax = 500000*niter//100
        else:
            xmax = int(counter * (niter+1)/len(fv_list))
        return xmax

    def get_progress(self):
        niter = self.num_iter.get()
        fv_array = self.canvas.demo_info[0]
        return fv_array.shape[0]/niter

    def read_callback_txt(self, cb_file):
        from .StateSequence import read_callback_txt_impl

        fv_list, x_list = read_callback_txt_impl(cb_file)
        if len(fv_list) == 0:
            # note that the first record is always included in these lists
            self.logger.info("making up the first record in callback.txt from the caller.")
            fv_array, x_array = self.state_info.get_init_info()[0:2]
            fv_list.append(fv_array[0])
            x_list.append(x_array[0])

        return fv_list, x_list

    def curr_folder_tracer(self, *args):
        job_name = self.curr_folder.get()
        i = self.folders.index(job_name)
        print("job_name=", job_name, i)
        self.curr_job = i
        job_state = self.job_list[i][1]
        self.job_state.set(job_state)
        state = Tk.NORMAL if job_state == JOB_STATES[0] else Tk.DISABLED
        for w in self.job_state_dependents:
            w.config(state=state)
        self.seed.set(self.job_list[i][5])

        w = self.stop_button
        if job_state == JOB_STATES[1]:
            w.config(fg='white', bg='red', state=Tk.NORMAL)
        else:
            configs = self.stop_button_configs
            w.config(fg=configs[0], bg=configs[1], state=Tk.DISABLED)

        self.folder_box.selection_clear()

    def maxnum_trials_tracer(self, *args):
        try:
            num_remaining = self.maxnum_trials.get() - self.num_jobs.get()
            state = Tk.NORMAL if num_remaining > 0 else Tk.DISABLED
            for w in self.remaining_dependents:
                w.config(state=state)
        except:
            # may be the spinbox is space
            pass

        self.update_try_model_composing_cb()

    def update_try_model_composing_cb(self):
        name = self.skip_button.cget('text')
        state = self.skip_button.cget('state')
        print("update_try_model_composing_cb: ", name, state)
        if name == 'Resume' and state == Tk.NORMAL:
            state = Tk.NORMAL
        else:
            state = Tk.DISABLED
        self.try_model_composing_cb.config(state=state)

    def terminate(self, immediately=False):
        if immediately:
            reply = True
        else:
            remaining_phrase = "" if self.maxnum_trials.get() == self.num_jobs.get() else " and stop the trials"
            reply = MessageBox.askokcancel("Terminate Confirmation",
                        "Are you sure to terminate the running job%s?" % remaining_phrase,
                        parent=self,
                        )
        if reply:
            self.runner.terminate()
            self.terminating = True

    def skip_to_next_trial(self, restart_params=None):
        num_remaining = self.maxnum_trials.get() - self.num_jobs.get()
        assert num_remaining > 0

        abort_phrase = "" if (self.terminating or self.inspecting) else " abort the running job and"
        skip_phrase = "resume from the next" if self.inspecting else "skip to the next"
        reply = MessageBox.askokcancel("Skip Confirmation",
                    "Are you sure to%s %s trial?" % (abort_phrase, skip_phrase),
                    parent=self,
                    )
        if reply:
            if self.terminating or self.inspecting:
                self.terminating = False
                self.start_next_trial(restart_params=restart_params)
            else:
                self.runner.terminate()

    def on_entry_root_folder(self):
        print("on_entry_root_folder")
        work_folder = os.path.join(self.optjob_folder.get(), '000')
        if os.path.exists(work_folder):
            self.canvas.get_ready_from_folder(work_folder)
            self.update_information(work_folder=work_folder)

    def set_inspection_mode(self):
        self.job_state.set(JOB_STATES[2])
        for w in self.remaining_dependents:
            w.config(state=Tk.DISABLED)

    def append_result(self, last_info):
        fv_vector = last_info[0][:,1]
        k = np.argmin(fv_vector)
        fv = fv_vector[k]
        if self.known_best_fv is None or fv < self.known_best_fv:
            self.known_best_fv = fv
            self.known_best_index = len(self.result_list)
            self.logger.info("updated known_best_index to %d with known_best_fv=%g", self.known_best_index, fv)

        x_array = last_info[1]
        result_info = OptJobResultInfo(fv=fv, params=x_array[k])
        self.result_list.append(result_info)
        self.logger.info("appended to result_list[%d]: fv=%g", len(self.result_list)-1, fv)

    def get_known_best_params(self, init_params):
        if len(self.result_list) == 0:
            ret_params = init_params
            self.logger.info("got best params from the initialization: num_params=%d", len(ret_params))
        else:
            best_result = self.result_list[self.known_best_index]
            ret_params = best_result.params
            self.logger.info("got best params from the %s result: num_params=%d", ordinal_str(self.known_best_index), len(ret_params))
        return ret_params

    def export(self, debug=True):
        if debug:
            from importlib import reload
            import Optimizer.LrfExporter
            reload(Optimizer.LrfExporter)
        from .LrfExporter import LrfExporter
        import molass_legacy.KekLib.CustomMessageBox as MessageBox

        optinit_info = self.optinit_info
        sd = optinit_info.sd
        dsets = optinit_info.dsets
        fullopt = self.fullopt
        canvas = self.canvas
        params = canvas.demo_info[1][canvas.curr_index]
        try:
            exporter = LrfExporter(self.fullopt, params, dsets)
            folder = exporter.export()
            fig_file = os.path.join(folder, "result_fig.jpg")
            self.save_the_result_figure(fig_file=fig_file)
            MessageBox.showinfo(
                    'Success Notification',
                    'Successfully exported to "%s"' % folder,
                    parent=self,
                    )
        except Exception as exc:
            log_exception(self.logger, "export: ")
            MessageBox.showerror(
                    'Error Notification',
                    'Failed to export due to: %s' % str(exc),
                    parent=self,
                    )

    def modify_restart(self, debug=True):
        if debug:
            from importlib import reload
            import Optimizer.ModifyStateDialog
            reload(Optimizer.ModifyStateDialog)
        from .ModifyStateDialog import ModifyStateDialog

        # self.grab_set()  # temporary fix to the grab_release problem

        try:
            mdialog = ModifyStateDialog(self.parent, self)
            mdialog.show()
            if mdialog.applied:
                pe_proxy = mdialog.get_pe_proxy()
                pdialog = self.parent_dialog
                v2_menu = pdialog.get_v2_menu()
                pdialog.after(500, lambda: v2_menu.show_peak_editor(pe_proxy=pe_proxy))
                self.cancel()
        except:
            from molass_legacy.KekLib.ExceptionTracebacker import log_exception
            log_exception(self.logger, "modify_restart: ", n=10)

        # self.grab_set()  # temporary fix to the grab_release problem

    def devel_test_back(self):
        from importlib import reload
        import Theory.PdbCrysolRoute
        reload(Theory.PdbCrysolRoute)
        from Theory.PdbCrysolRoute import demo
        demo(self)

    def devel_test(self):
        print("devel_test")
        from importlib import reload
        import Tools.EmbedCushion
        reload(Tools.EmbedCushion)
        from molass_legacy.Tools.EmbedCushion import embed_cushion

        embed_cushion(self)



# coding: utf-8
"""
    SecTheory.SinglePoreNonCf.py

    Copyright (c) 2022, SAXS Team, KEK-PF
"""
import numpy as np
from scipy.optimize import basinhopping
import molass_legacy.KekLib.DebugPlot as plt
from DataUtils import get_in_folder
from SecTheory.BasicModels import single_pore_elution

def single_pore_pdf(x, rg, rp, x0, nperm, tperm, me, mp):
    return single_pore_elution(x - x0, rg, rp, nperm, tperm, me, mp)

class SinglePoreSolver:
    def __init__(self, x, y, n_species, rg_curve=None):
        self.x = x
        self.y = y
        self.area = np.sum(y)
        self.n_species = n_species
        self.rg_curve = rg_curve
        if rg_curve is not None:
            self.mask = rg_curve.get_mask()
            self.xm, self.ym, self.valid_rg = rg_curve.get_valid_curves()
            self.rg_area = np.sum(self.valid_rg)
            print("area=", self.area, "rg_area=", self.rg_area)

    def split_params(self, params):
        sep1 = 4
        sep2 = sep1 + self.n_species
        rp_params = params[0:sep1]
        rg_params = params[sep1:sep2]
        wt_params = params[sep2:]
        return rp_params, rg_params, wt_params

    def objective_function(self, params, debug=False):
        x = self.x
        y = self.y
        rp_params, rg_params, wt_params = self.split_params(params)

        ty = np.zeros(len(y))
        penalty = 0
        tw = 0
        last_rg = None
        cy_list = []
        for k, rg in enumerate(rg_params):
            if k < len(wt_params):
                weight = wt_params[k]
            else:
                weight = 1 - tw
            tw += weight
            cy = weight*self.area*single_pore_pdf(x, rg, *rp_params, 1, 1)
            cy_list.append(cy)
            ty += cy
            penalty += min(0, weight - 1e-3)**2
            if last_rg is not None:
                penalty += min(0, last_rg - rg)**2
            last_rg = rg

        elution_dev = np.log10(np.sum((ty - y)**2)/self.area)/2

        valid_rg = self.valid_rg
        t_rg = np.zeros(len(valid_rg))
        tym = ty[self.mask]
        for r, xr_cy in zip(rg_params, cy_list):
            t_rg += r * xr_cy[self.mask]/tym

        rg_dev = np.log10(np.sum((t_rg -valid_rg)**2)/self.rg_area)/2   # equivalent to np.log10(np.sqrt(...

        fv = elution_dev + rg_dev + penalty*1e8
        if debug:
            print("penalty=", penalty)
            print("fv=", fv)
            in_folder = get_in_folder()
            rg_estimation = np.zeros(len(y))
            with plt.Dp():
                fig, ax = plt.subplots()
                ax.set_title("Single Pore Model Fit for %s FV=%.3g" % (in_folder, fv), fontsize=20)
                ax.plot(x, y, color="orange", label="Data")
                tw = 0
                for k, cy in enumerate(cy_list):
                    if k < len(wt_params):
                        weight = wt_params[k]
                    else:
                        weight = 1 - tw
                    rg = rg_params[k]
                    rg_estimation += rg*cy/ty
                    print([k], "rg=%.3g, weight=%.3g" % (rg, weight))
                    tw += weight
                    ax.plot(x, cy, ":", lw=2, label="Rg=%.3g" % rg)
                ax.plot(x, ty, ":", lw=2, color="red", label="Model Total")
                xmin, xmax = ax.get_xlim()
                ymin, ymax = ax.get_ylim()
                ymax = max(np.max(y), np.max(ty))*1.2
                ax.set_ylim(ymin, ymax)

                x0 = rp_params[1]
                ax.plot([x0, x0], [ymin, ymax], color="yellow", label="x0")
                ax.legend()

                w = 0.05
                tx = xmin*(1-w) + xmax*w
                w = 0.9
                ty = ymin*(1-w) + ymax*w
                ax.text(tx, ty, "Pore Size=%.3g" % rp_params[0], fontsize=30, alpha=0.2)
                for name, v in zip(["rp", "x0", "nperm", "tperm"], rp_params):
                    print("%s = %.3g" % (name, v))
                print("fv=%g" % fv)
                print("elution_dev=%g" % elution_dev, "rg_dev=%g" % rg_dev, "penalty=%g" % penalty)
                axt = ax.twinx()
                axt.grid(False)
                if self.rg_curve is not None:
                    for k, (x_, y_, rg_) in enumerate(self.rg_curve.get_curve_segments()):
                        label = 'Rg (guinier)' if k == 0 else None
                        axt.plot(x_, rg_, color='gray', alpha=0.5, label=label)

                axt.plot(x, rg_estimation, ":", lw=2, color="gray", label="Rg (model fit)")
                axt.set_ylim(10, rg_params[0]*1.2)
                axt.legend(loc="center right")
                ret = plt.show()
                if not ret:
                    debug = False
        return fv

    def optimize(self, init_params, n_iters=100):
        ret = basinhopping(self.objective_function, init_params, niter=n_iters)
        return ret

def fit_single_pore_pdf(ecurve, rg_curve, init_params, n_species=3, n_iters=100, n_trials=0):
    import warnings
    warnings.filterwarnings("ignore")

    x = ecurve.x
    y = ecurve.y

    solver = SinglePoreSolver(x, y, n_species, rg_curve=rg_curve)

    ret = None
    best_fv = None
    for k in range(n_trials):
        print([k], "optimizing ...")
        ret = solver.optimize(init_params, n_iters=n_iters)
        if best_fv is None or ret.fun < best_fv:
            best_fv = ret.fun
            print([k], "best_fv=%g" % best_fv)
            init_params = ret.x

    solver.objective_function(init_params, debug=True)
